//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop


#include "Global2.h"      // WM_SHOWSTATUS
#include "PGPsdk6.h"
#include "PGPsdk6Calls.h"

// higher level
PGPError __fastcall OpenKeyRings(const AnsiString &pubring, const AnsiString &secring, PGPKeySetRef *keySet);
PGPError __fastcall OpenKeyKeyRingsMutable(const AnsiString &pubring, const AnsiString &secring, PGPKeySetRef *keySet);
PGPError __fastcall OpenKeyRings(const AnsiString &pubring, const AnsiString &secring,
   PGPKeyRingOpenFlags options, PGPKeySetRef *keySet);
PGPError __fastcall OpenKeyRing(const AnsiString &pubring, PGPKeySetRef *keySet);
PGPError __fastcall OpenImportKeyRing(const AnsiString &pubring, PGPKeySetRef *keySet);
// cipher
PGPError __fastcall EncryptText(const AnsiString &pubring, const AnsiString &secring, TStrings *keyIDs,
   const AnsiString &signKeyID, const AnsiString &signPassphrase, TStrings *Text);
PGPError __fastcall EncryptText(const AnsiString &passphrase, TStrings *Text);
// key for list view
PGPError __fastcall GetKey(PGPKeyRef KeyRef, TStrings *Key);
PGPError __fastcall GetSubKey(PGPKeyIterRef keyDBIter, TStrings *Key);
PGPError __fastcall GetSubKey(PGPKeyRef theKey, TStrings *Key);
PGPError __fastcall GetKeyComplete(PGPKeyRef theKey, TStrings *Key);
PGPError __fastcall GetKey(PGPKeySetRef &keySet, const AnsiString &KeyIDStr, PGPKeyRef *theKey,
   PGPPublicKeyAlgorithm PreferredAlgo = kPGPPublicKeyAlgorithm_Invalid);
AnsiString __fastcall GetPreferredKey(TStrings *keyids, const AnsiString &preferredalgo);
// keyids
PGPError __fastcall GetKeyID(const AnsiString &Pubring, const AnsiString &UserId, TStrings* KeyIDs,
   const AnsiString &PreferredAlgo);
PGPError __fastcall GetKeyIDs(const AnsiString &Pubring, TStrings* UserIds, TStrings* KeyIDs,
   const AnsiString &PreferredAlgo);
// miscellaneous
PGPError __fastcall ExportKeySet(PGPKeySetRef keys, PGPExportFormat ExportFormat,
   bool IncludePrivates, TStrings *exports);
PGPError __fastcall GetDisableKeyIds(PGPKeySetRef keyset, TStrings *keyids);
PGPError __fastcall GetRevokedKeyIds(PGPKeySetRef keyset, TStrings *keyids);
void __fastcall AddKeyToKeys(TStrings *key, TStrings *keys);
void __fastcall UpdateDescriptionField(TStrings *key);
// support
AnsiString __fastcall PublicKeyAlgoStr(PGPInt32 algo);
AnsiString __fastcall StatusStr(int validity, PGPBoolean revoked, PGPBoolean expired, PGPBoolean disabled);
AnsiString __fastcall AlgorithmStr(int algo);
AnsiString __fastcall ValidityStr(int val);

#pragma package(smart_init)
//---------------------------------------------------------------------------
// Init PGP6
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRings(const AnsiString &pubring, const AnsiString &secring,
   PGPKeySetRef *keySet)
{
   return OpenKeyRings(pubring, secring, kPGPKeyRingOpenFlags_None, keySet);
}
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRingsMutable(const AnsiString &pubring, const AnsiString &secring,
   PGPKeySetRef *keySet)
{
   return OpenKeyRings(pubring, secring, kPGPKeyRingOpenFlags_Mutable, keySet);
}
//---------------------------------------------------------------------------

/*PGPError __fastcall OpenKeySet(const AnsiString &pubring, const AnsiString &secring,
   PGPKeySetRef *keySet)
{
   assert(FileExists(pubring) && FileExists(secring));

   err6 = kPGPError_NoErr;
   keySet = kInvalidPGPKeySetRef;
   PGPFileSpecRef pubFileRef = kInvalidPGPFileSpecRef;
   PGPFileSpecRef secFileRef = kInvalidPGPFileSpecRef;

   err6 = PGPNewFileSpecFromFullPath(context, pubring.c_str(), &pubFileRef);
   if(IsPgpError(err6, "Unable to get PGP public keyring path!"))
      goto exit;

   err6 = PGPNewFileSpecFromFullPath(context, secring.c_str(), &secFileRef);
   if (IsPgpError(err6, "Unable to get PGP secret keyring path!"))
      goto exit;

   err6 = PGPOpenKeyRingPair(context, kPGPKeyRingOpenFlags_none, pubFileRef, secFileRef, keySet);
   if (IsPgpError(err6, "Failed to open QS PGP keyrings"))
      goto exit;

exit:

   if (PGPFileSpecRefIsValid(pubFileRef))
      PGPFreeFileSpec(pubFileRef);

   if (PGPFileSpecRefIsValid(secFileRef))
      PGPFreeFileSpec(secFileRef);

   return err6;
}*/
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRings(const AnsiString &pubring, const AnsiString &secring,
   PGPKeyRingOpenFlags flags, PGPKeySetRef *keySet)
{
   assert(FileExists(pubring) && FileExists(secring));

  *keySet = kInvalidPGPKeySetRef;
   PGPFileSpecRef pubFileRef = kInvalidPGPFileSpecRef;
   PGPFileSpecRef secFileRef = kInvalidPGPFileSpecRef;

   err6 = PGPNewFileSpecFromFullPath(context, pubring.c_str(), &pubFileRef);
   if(IsPgpError(err6, "Unable to get PGP public keyring path!"))
      goto exit;

   err6 = PGPNewFileSpecFromFullPath(context, secring.c_str(), &secFileRef);
   if (IsPgpError(err6, "Unable to get PGP secret keyring path!"))
      goto exit;

   err6 = PGPOpenKeyRingPair(context, flags, pubFileRef, secFileRef, keySet);
   if (IsPgpError(err6, "Failed to open QS PGP keyrings"))
      goto exit;

exit:

   if (PGPFileSpecRefIsValid(pubFileRef))
      PGPFreeFileSpec(pubFileRef);

   if (PGPFileSpecRefIsValid(secFileRef))
      PGPFreeFileSpec(secFileRef);

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRing(const AnsiString &pubring, PGPKeySetRef *keySet)
{
   assert(FileExists(pubring));

   if (ExtractFileExt(pubring).LowerCase() == ".asc")
      return  OpenImportKeyRing(pubring, keySet);

   err6 = kPGPError_NoErr;
   keySet = kInvalidPGPKeySetRef;
   PGPFileSpecRef pubFileRef = kInvalidPGPFileSpecRef;

   err6 = PGPNewFileSpecFromFullPath(context, pubring.c_str(), &pubFileRef);
   if(IsPgpError(err6, "Unable to get PGP public keyring path!"))
      goto exit;

   err6 = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_None, pubFileRef, keySet);
   if (IsPgpError(err6, "Failed to open QS PGP keyring"))
      goto exit;

exit:

   if (PGPFileSpecRefIsValid(pubFileRef))
      PGPFreeFileSpec(pubFileRef);

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall OpenImportKeyRing(const AnsiString &pubring, PGPKeySetRef *keySet)
{
   PGPFileSpecRef pubFileRef = kInvalidPGPFileSpecRef;

   err6 = PGPNewFileSpecFromFullPath( context, pubring.c_str(), &pubFileRef);
   if(IsPGPError(err6))
      goto exit;

   err6 = PGPImportKeySet(context, keySet, PGPOInputFile(context, pubFileRef), PGPOLastOption(context),
      0, 0, 0, 0);
   if (IsPGPError(err6))
      goto exit;

exit:

   if (PGPFileSpecRefIsValid(pubFileRef))
      PGPFreeFileSpec(pubFileRef);

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetSubKey(PGPKeyRef theKey, TStrings *Key)
{
   assert(theKey != NULL && Key != NULL);

   PGPInt32 bits, algo;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyListRef keyList = kInvalidPGPKeyListRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPSubKeyRef subKey = kInvalidPGPSubKeyRef;
   PGPKeyID keyID;

   err6 = PGPNewSingletonKeySet(theKey, &keySet);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPOrderKeySet(keySet, kPGPAnyOrdering, &keyList);
   if (IsPgpError(err6, "Unable to create ordered key list!"))
      goto exit;

   err6 = PGPNewKeyIter(keyList, &keyIter);
   if (IsPgpError(err6, "Unable to create keylist iterator!"))
      goto exit;

   err6 = PGPKeyIterSubKey(keyIter, &subKey);
   if (IsPGPError(err6))
      goto exit;

   // keyid
   err6 = PGPGetKeyIDFromSubKey(subKey, &keyID);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
   if (IsntPGPError(err6)) Key->Add("subkeyid=" + AnsiString(keyidstr));
      else goto exit;

   // bits
   err6 = PGPGetSubKeyNumber(subKey, kPGPKeyPropBits, &bits);
   if (IsntPGPError(err6)) Key->Add("subbits=" + IntToStr(bits));
      else goto exit;

   // algorithm
   err6 = PGPGetSubKeyNumber(subKey, kPGPKeyPropAlgID, &algo);
   if (IsntPGPError(err6)) Key->Add("subalgo=" + IntToStr(algo));
      else goto exit;

exit:

   if (err6 == kPGPError_EndOfIteration)
      err6 = kPGPError_NoErr;

   if (IsPgpError(err6, "Failed to get requested subkey!"));

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeyListRefIsValid(keyList))
      PGPFreeKeyList(keyList);

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall EncryptText(const AnsiString &pubring, const AnsiString &secring,
   TStrings *keyIDs, const AnsiString &signKeyID,
   const AnsiString &signPassphrase,
   TStrings *Text)
{
   assert(Text != NULL && keyIDs != NULL);

   PGPKeyID keyID;
   bool gotencrypt = false, gotsign = false;
   AnsiString line, keyid, msgstr;
   void *data = NULL; PGPSize size;
   PGPKeySetRef secSet = kInvalidPGPKeySetRef;
   PGPKeySetRef pubSet = kInvalidPGPKeySetRef;
   PGPKeySetRef encSet = kInvalidPGPKeySetRef;
   PGPKeySetRef singleSet = kInvalidPGPKeySetRef;
   PGPKeyRef signKey = kInvalidPGPKeyRef;
   PGPKeyRef encKey = kInvalidPGPKeyRef;

   // create keyset and optionlist for text encryption
   if (!pubring.IsEmpty() && keyIDs != NULL && keyIDs->Count > 0)
   {
      err6 = OpenKeyRing(pubring, &pubSet);
      if (IsPGPError(err6))
         goto exit;

      err6 = PGPNewKeySet(context, &encSet);
      if (IsPGPError(err6))
         goto exit;

      int i = 0; while(i < keyIDs->Count)
      {
         keyid = keyIDs->Strings[i].SubString(3,8);

         err6 = PGPGetKeyIDFromString(keyid.c_str(), &keyID);
         if (IsPGPError(err6))
            goto exit;

         err6 = PGPGetKeyByKeyID(pubSet, &keyID, kPGPPublicKeyAlgorithm_Invalid, &encKey);
         if (IsPGPError(err6))
            goto exit;

         err6 = PGPNewSingletonKeySet(encKey, &singleSet);
         if (IsPGPError(err6))
            goto exit;

         err6 = PGPAddKeys(singleSet, encSet);
         if (IsPGPError(err6))
            goto exit;

         PGPFreeKeySet(singleSet);

         i++;
      }

      gotencrypt = true;
   }

   // get key and create optionlist for signing
   if (!secring.IsEmpty() && !signKeyID.IsEmpty())
   {
      err6 = OpenKeyRing(secring, &secSet);
      if (IsPGPError(err6))
         goto exit;

      keyid = signKeyID.SubString(3,8);

      err6 = PGPGetKeyIDFromString(keyid.c_str(), &keyID);
      if (IsPGPError(err6))
            goto exit;

      err6 = PGPGetKeyByKeyID(secSet, &keyID, kPGPPublicKeyAlgorithm_Invalid, &signKey);
      if (IsPGPError(err6))
         goto exit;

      gotsign = true;
   }

   if (gotencrypt && !gotsign)
   {
      err6 = PGPEncode( context,
         PGPOEncryptToKeySet( context, encSet),
         PGPOInputBuffer( context, Text->Text.c_str(), Text->Text.Length()),
         PGPOAllocatedOutputBuffer( context, &data, -1, &size),
         PGPOVersionString( context, "N/A" ),
         PGPOArmorOutput( context, true),
         PGPODataIsASCII( context, true),
         PGPOCompression( context, true),
         PGPOLastOption( context ),
         0,0,0,0,0,0,0,0);
      if (IsPgpError(err6, "Sorry, I failed to encrypt your text!"))
         goto exit;

      Text->SetText((char*)data);
      goto exit;
   }

   if (gotsign && !gotencrypt)
   {
      err6 = PGPEncode( context,
         PGPOSignWithKey( context, signKey,
            PGPOPassphrase( context, signPassphrase.c_str() ),
            PGPOLastOption( context ), 0),
         PGPOClearSign( context, true ),
         PGPOInputBuffer( context, Text->Text.c_str(), Text->Text.Length()),
         PGPOAllocatedOutputBuffer( context, &data, -1, &size),
         PGPOVersionString( context, "N/A" ),
         PGPOArmorOutput( context, true),
         PGPODataIsASCII( context, true),
         PGPOCompression( context, true),
         PGPOLastOption( context ),
         0,0,0,0,0,0,0);
      if (IsPgpError(err6, "Sorry, I failed to sign your text!"))
         goto exit;

      Text->SetText((char*)data);
      goto exit;
   }

   err6 = PGPEncode( context,
      PGPOEncryptToKeySet( context, encSet),
      PGPOSignWithKey( context, signKey,
         PGPOPassphrase( context, signPassphrase.c_str() ),
         PGPOLastOption( context ), 0),
      PGPOClearSign( context, false ),
      PGPOInputBuffer( context, Text->Text.c_str(), Text->Text.Length()),
      PGPOAllocatedOutputBuffer( context, &data, -1, &size),
      PGPOVersionString( context, "N/A" ),
      PGPOArmorOutput( context, true),
      PGPODataIsASCII( context, true),
      PGPOCompression( context, true),
      PGPOLastOption( context ),
      0,0,0,0,0,0);
   if (IsPgpError(err6, "Sorry, I failed to encrypt/sign your text!"))
      goto exit;

   Text->SetText((char*)data);

exit:

   if (PGPKeySetRefIsValid(singleSet))
      PGPFreeKeySet(singleSet);

   if (PGPKeySetRefIsValid(encSet))
      PGPFreeKeySet(encSet);

   if (PGPKeySetRefIsValid(pubSet))
      PGPFreeKeySet(pubSet);

   if (PGPKeySetRefIsValid(secSet))
      PGPFreeKeySet(secSet);

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall EncryptText(const AnsiString &Passphrase, TStrings *Text)
{
   AnsiString line, msgstr;
   void *data = NULL;
   PGPSize size;
   err6 = PGPEncode( context,
      PGPOConventionalEncrypt( context,
         PGPOPassphrase( context, Passphrase.c_str() ),
         PGPOLastOption( context ), 0),
      PGPOInputBuffer( context, Text->Text.c_str(), Text->Text.Length()),
      PGPOAllocatedOutputBuffer( context, &data, -1,&size),
      PGPOVersionString( context, "N/A" ),
      PGPOArmorOutput( context, true),
      PGPODataIsASCII( context, true),
      PGPOCompression( context, true),
      PGPOLastOption( context ),
      0,0,0,0,0,0,0,0
   );

   if (IsPGPError(err6))
      goto exit;

   Text->SetText((char*)data);

exit:

   if (data)
      PGPFreeData(data);

   return err6;
}
//---------------------------------------------------------------------------

AnsiString __fastcall PublicKeyAlgoStr(PGPInt32 algo)
{
   switch (algo)
   {
      case kPGPPublicKeyAlgorithm_Invalid:        return "INVALID";
      case kPGPPublicKeyAlgorithm_RSA:            return "RSA";
      case kPGPPublicKeyAlgorithm_RSAEncryptOnly: return "RSAE";
      case kPGPPublicKeyAlgorithm_RSASignOnly:    return "RSAS";
      case kPGPPublicKeyAlgorithm_ElGamal:	  return "DH";
      case kPGPPublicKeyAlgorithm_DSA:		  return "DSA";
   }
   return "UNKNOWN";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GetPreferredKey(TStrings *keyids, const AnsiString &preferredalgo)
{
 //changed this assertion to dialog (in GetKeyIDs below) explaining problem of no key found
 //assert(keyids->Count > 0);

   int index = 0;
   AnsiString key;
   for (int i = 0; i < keyids->Count; i++)
   {
      if (keyids->Strings[i].Pos(" " + preferredalgo))
      {
         index = i;
         break;
      }
   }

   if (keyids->Count > 0)
      key = keyids->Strings[index];

   return key;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKeyID(const AnsiString &Pubring, const AnsiString &UserID,
   TStrings* KeyIDs, const AnsiString &PreferredAlgo)
{
   assert(FileExists(Pubring) && UserID != "" && KeyIDs != NULL);

   TStringList *userids = new TStringList();
   userids->Add(UserID);

   err6 = GetKeyIDs(Pubring, userids, KeyIDs, PreferredAlgo);

   delete userids;

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKeyIDs(const AnsiString &pubring, TStrings *UserIDs,
   TStrings* KeyIDs, const AnsiString &PreferredAlgo)
{
   assert(FileExists(pubring) && UserIDs != NULL && UserIDs->Count > 0 && KeyIDs != NULL);

   KeyIDs->Clear();
   TStringList *keyids = new TStringList();
   AnsiString chosenkey;

   char keyIDstr[kPGPMaxKeyIDStringSize];
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeySetRef filterSet = kInvalidPGPKeySetRef;
   PGPKeyListRef keyList = kInvalidPGPKeyListRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyRef theKey = kInvalidPGPKeyRef;
   PGPFilterRef keyFilter = kInvalidPGPFilterRef;
   PGPInt32 pkalg;
   PGPKeyID keyID;

   err6 = OpenKeyRing(pubring, &keySet);
   if (IsPGPError(err6))
      goto exit;

   for (int i = 0; i < UserIDs->Count; i++)
   {
      AnsiString userid = UserIDs->Strings[i];

      err6 = PGPNewUserIDEmailFilter(context, userid.c_str(), kPGPMatchSubString, &keyFilter);
      if(IsPGPError(err6))
         goto exit;

      err6 = PGPFilterKeySet(keySet, keyFilter, &filterSet);
      if(IsPGPError(err6))
         goto exit;

      err6 = PGPOrderKeySet(filterSet, kPGPAnyOrdering, &keyList);
      if(IsPGPError(err6))
         goto exit;

      err6 = PGPNewKeyIter(keyList, &keyIter);
      if (IsPGPError(err6))
         goto exit;

      keyids->Clear();
      while (PGPKeyIterNext(keyIter, &theKey) != kPGPError_EndOfIteration)
      {
         err6 = PGPGetKeyIDFromKey(theKey, &keyID);
         if (IsPGPError(err6))
            goto exit;

         err6 = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyIDstr);
         if (IsPGPError(err6))
            goto exit;

         err6 = PGPGetKeyNumber(theKey, kPGPKeyPropAlgID, &pkalg);
         if (IsPGPError(err6))
            goto exit;

         keyids->Add(AnsiString(keyIDstr) + " " + PublicKeyAlgoStr(pkalg));
      }

      // should be no empty keyids but check anyway
      int i = 0; while (i < keyids->Count)
      {
         if (keyids->Strings[i].Trim().IsEmpty())
            keyids->Delete(i);
         else
            i++;
      }

      chosenkey = GetPreferredKey(keyids, PreferredAlgo);
      if (chosenkey.IsEmpty())
      {
         AnsiString ErrMsg =
           "I have a problem. There does not seem to be PGP key(s) for:    " +
            UserIDs->Text + "\nI'm quitting so you can take a look.";

         Application->MessageBox(ErrMsg.c_str(), "Error", MB_OK);
         err6 = kPGPError_ItemWasDeleted;
         goto exit;
      }

      KeyIDs->Add(chosenkey);

      if (PGPKeyIterRefIsValid(keyIter))
         PGPFreeKeyIter(keyIter);

      if (PGPKeyListRefIsValid(keyList))
         PGPFreeKeyList(keyList);

      if (PGPFilterRefIsValid(keyFilter))
         PGPFreeFilter(keyFilter);
   }

exit:

   if (PGPFilterRefIsValid(keyFilter))
      PGPFreeFilter(keyFilter);

   if (PGPKeyListRefIsValid(keyList))
      PGPFreeKeyList(keyList);

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   delete keyids;

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall ExportKeySet(PGPKeySetRef keySet, PGPExportFormat ExportFormat,
   bool IncludePrivates, TStrings *exports)
{
   assert(PGPKeySetRefIsValid(keySet) && exports != NULL);

   exports->Clear();
   void *data=NULL; PGPSize size;

   err6 = PGPExportKeySet(keySet,
      PGPOAllocatedOutputBuffer(context, &data, -1, &size), // -1 = no size constraint
      PGPOExportFormat(context, ExportFormat),
      PGPOExportPrivateKeys(context, IncludePrivates),
      PGPOVersionString(context, "N/A"),
      PGPOLastOption(context)
   );
   if (IsPgpError(err6,"Failed to export keys!"))
      goto exit;

   exports->SetText((char*)data);

exit:

   if (data) PGPFreeData(data);
   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetDisableKeyIds(PGPKeySetRef keyset, TStrings *keyids)
{
   assert(PGPKeySetRefIsValid(keyset) && keyids != NULL);

   keyids->Clear();
   PGPKeyRef key = kInvalidPGPKeyRef;
   PGPFilterRef filter = kInvalidPGPFilterRef;
   PGPKeySetRef diskeys = kInvalidPGPKeySetRef;
   PGPKeyListRef diskeyslist = kInvalidPGPKeyListRef;
   PGPKeyIterRef listIterator = kInvalidPGPKeyIterRef;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeyID keyid;

   err6 = PGPNewKeyDisabledFilter(context, true, &filter);
   if (IsPgpError(err6,"Failed to create disabled key filter."))
      goto exit;

   err6 = PGPFilterKeySet(keyset,filter,&diskeys);
   if (IsPgpError(err6,"Failed to get disabled keys."))
      goto exit;

   err6 = PGPOrderKeySet(diskeys,kPGPAnyOrdering,&diskeyslist);
   if (IsPgpError(err6,"Failed to create disabled key list."))
      goto exit;

   err6 = PGPNewKeyIter(diskeyslist,&listIterator);
   if (IsPgpError(err6,"Failed to create key list iterator."))
      goto exit;

   while (IsntPGPError(err6 = PGPKeyIterNext(listIterator, &key)) && key)
   {
      err6 = PGPGetKeyIDFromKey(key, &keyid);
      if (IsPgpError(err6, "Failed to get key id from revoked key."))
         goto exit;

      err6 = PGPGetKeyIDString(&keyid, kPGPKeyIDString_Abbreviated, keyidstr);
      if (IsPgpError(err6,"Failed to get key id string!"))
         goto exit;

      keyids->Add(keyidstr);
   }
   if (err6 == kPGPError_EndOfIteration)
      err6 = kPGPError_NoErr;

   if (IsPgpError(err6, "Failed to complete iteration of revoked key list."))
      goto exit;

exit:
   if (PGPKeyIterRefIsValid(listIterator))
      PGPFreeKeyIter(listIterator);

   if (PGPKeyListRefIsValid(diskeyslist))
      PGPFreeKeyList(diskeyslist);

   if (PGPKeySetRefIsValid(diskeys))
      PGPFreeKeySet(diskeys);

   if (PGPFilterRefIsValid(filter))
      PGPFreeFilter(filter);

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetRevokedKeyIds(PGPKeySetRef keyset, TStrings *keyids)
{
   assert(PGPKeySetRefIsValid(keyset) && keyids != NULL);

   keyids->Clear();
   PGPKeyRef key = kInvalidPGPKeyRef;
   PGPFilterRef filter = kInvalidPGPFilterRef;
   PGPKeySetRef revkeys = kInvalidPGPKeySetRef;
   PGPKeyListRef revkeyslist = kInvalidPGPKeyListRef;
   PGPKeyIterRef listIterator = kInvalidPGPKeyIterRef;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeyID keyid;

   err6 = PGPNewKeyRevokedFilter(context,true,&filter);
   if (IsPgpError(err6,"Failed to create revoked key filter."))
      goto exit;

   err6 = PGPFilterKeySet(keyset,filter,&revkeys);
   if (IsPgpError(err6,"Failed to get revoked keys."))
      goto exit;

   err6 = PGPOrderKeySet(revkeys,kPGPAnyOrdering,&revkeyslist);
   if (IsPgpError(err6,"Failed to create revoked key list."))
      goto exit;

   err6 = PGPNewKeyIter(revkeyslist, &listIterator);
   if (IsPgpError(err6,"Failed to create key list iterator."))
      goto exit;

   while (IsntPGPError(err6=PGPKeyIterNext(listIterator,&key)) && key)
   {
      err6 = PGPGetKeyIDFromKey(key,&keyid);
      if (IsPgpError(err6,"Failed to get key id from revoked key."))
         goto exit;

      err6 = PGPGetKeyIDString(&keyid,kPGPKeyIDString_Abbreviated,keyidstr);
      if (IsPgpError(err6,"Failed to get key id string!"))
         goto exit;

      keyids->Add(keyidstr);
   }
   if (err6 == kPGPError_EndOfIteration)
      err6 = kPGPError_NoErr;

   if (IsPgpError(err6, "Failed to complete iteration of revoked key list"))
      goto exit;

exit:
   if (PGPKeyIterRefIsValid(listIterator))
      PGPFreeKeyIter(listIterator);

   if (PGPKeyListRefIsValid(revkeyslist))
      PGPFreeKeyList(revkeyslist);

   if (PGPKeySetRefIsValid(revkeys))
      PGPFreeKeySet(revkeys);

   if (PGPFilterRefIsValid(filter))
      PGPFreeFilter(filter);

   return err6;
}
//---------------------------------------------------------------------------

/*PGPError __fastcall OpenKeyRings(PGPContextRef context, PGPKeyRingOpenFlags options,
   const AnsiString pubring, const AnsiString secring, PGPKeySetRef &keySet)
{
   assert(PGPContextRefIsValid(context) && PGPKeySetRefIsValid(keySet));

   keySet = kInvalidPGPKeySetRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;

   err6 = PGPNewFileSpecFromFullPath(context, pubring.c_str(), &PubKeyRing);
   if(IsPgpError(err6,"Failed to get QS PGP public keyring file spec."))
      goto exit;

   err6 = PGPNewFileSpecFromFullPath(context, secring.c_str(), &SecKeyRing);
   if(IsPgpError(err6,"Failed to get QS PGP secret keyring file spec."))
      goto exit;

   if (pubring != "" && secring != "")
   {
      err6 = PGPOpenKeyRingPair(context, false, PubKeyRing, SecKeyRing, &keySet);
      if (IsPgpError(err6,"Failed to open QS PGP keyrings"))
         goto exit;
   }

   if (pubring != "")
   {
      err6 = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_Mutable, PubKeyRing, &keySet);
      if (IsPgpError(err6,"Failed to open PGP public keyring"))
         goto exit;
   }

   if (secring != "")
   {
      err6 = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_Mutable, SecKeyRing, &keySet);
      if (IsPgpError(err6,"Failed to open PGP secret keyring"))
         goto exit;
   }

exit:

   if (PGPFileSpecRefIsValid(SecKeyRing))
      PGPFreeFileSpec(SecKeyRing);

   if (PGPFileSpecRefIsValid(PubKeyRing))
      PGPFreeFileSpec(PubKeyRing);

   return err6;
}*/
//---------------------------------------------------------------------------

void __fastcall AddKeyToKeys(TStrings *key, TStrings *keys)
{
   assert(key != NULL && keys != NULL);

   AnsiString keystr;

   for (int i = 0; i < key->Count; i++)
   {
      if (i > 0 && i < key->Count)
         keystr += "\255";
      keystr += key->Strings[i];
   }

   keys->Add(keystr);
}
//---------------------------------------------------------------------------

void __fastcall UpdateDescriptionField(TStrings *key)
{
   assert(key != NULL);

   bool secret = key->Strings[lvSecret] == "1";
   AnsiString descrip = key->Strings[lvAlgo];

   //if (descrip == "DSS") descrip = "DH/DSS";

   if (descrip == "RSA") descrip += (secret ? " legacy key pair" : " legacy public key"); else
   if (descrip == "DH/DSS") descrip += (secret ? " key pair" : " public key"); else
   if (descrip == "RSA/RSA") descrip = AnsiString("RSA") +
      (secret ? " key pair" : " public key");

   if (descrip != key->Strings[lvAlgo])
      key->Strings[lvAlgo] = descrip;
}
//---------------------------------------------------------------------------

AnsiString __fastcall StatusStr(int validity, PGPBoolean revoked, PGPBoolean expired,
   PGPBoolean disabled)
{
   if (revoked ) return "revoked";
   if (expired ) return "expired";
   if (disabled) return "disabled";

   return ValidityStr(validity);
}
//---------------------------------------------------------------------------

AnsiString __fastcall AlgorithmStr(int algo)
{
   switch (algo)
   {
      case kPGPPublicKeyAlgorithm_RSA: return "RSA";
      case kPGPPublicKeyAlgorithm_RSAEncryptOnly: return "RSA(e)";
      case kPGPPublicKeyAlgorithm_RSASignOnly: return "RSA(s)";
      case kPGPPublicKeyAlgorithm_ElGamal: return "DH";
      case kPGPPublicKeyAlgorithm_DSA: return "DSS";
      //case kPGPPublicKeyAlgorithm_ECEncrypt: return "EC(e)";
      //case kPGPPublicKeyAlgorithm_ECSign: return "EC(s)";
   }
   return "unknown";
}
//---------------------------------------------------------------------------

AnsiString __fastcall ValidityStr(int val)
{
   switch (val)
   {
      case kPGPValidity_Unknown : return "";
      case kPGPValidity_Invalid : return "invalid";
      case kPGPValidity_Marginal: return "marginal";
      case kPGPValidity_Complete: return "ok";
   }
   return "unknown";
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKey(PGPKeySetRef &keySet, const AnsiString &KeyIDStr,
   PGPKeyRef *theKey, PGPPublicKeyAlgorithm PreferredAlgo)
{
   assert(!KeyIDStr.IsEmpty() && PGPKeySetRefIsValid(keySet));

   PGPKeyID keyID;

   err6 = PGPGetKeyIDFromString(KeyIDStr.c_str(), &keyID);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyByKeyID(keySet, &keyID, PreferredAlgo, theKey);
   if (IsPgpError(err6, "Unable to get the required key."))
      goto exit;
exit:

   if (IsPgpError(err6, "Failed to retrieve the requested key."));

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKey(PGPKeyRef KeyRef, TStrings *Key)
{
   assert(KeyRef != NULL && Key != NULL);

   char userid[kPGPMaxUserIDSize];
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPBoolean revoked, expired, disabled, secret;
   PGPInt32 bits, validity, algo;
   PGPUInt32 adks, useridsize;
   PGPKeyID keyID;

   Key->Clear();

   // 0 userid  lvUserId
   err6 = PGPGetPrimaryUserIDNameBuffer(KeyRef, kPGPMaxUserIDSize - 1, userid, &useridsize);
   Key->Add(userid);

   // 1 keyid  lvKeyID
   err6 = PGPGetKeyIDFromKey(KeyRef, &keyID);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
   if (IsPGPError(err6))
      goto exit;

   Key->Add(keyidstr);

   // 2 status  lvStatus
   err6 = PGPGetKeyBoolean(KeyRef, kPGPKeyPropIsRevoked, &revoked);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyBoolean(KeyRef, kPGPKeyPropIsDisabled, &disabled);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyBoolean(KeyRef, kPGPKeyPropIsExpired, &expired);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyNumber(KeyRef, kPGPKeyPropValidity, &validity);
   if (IsntPGPError(err6)) Key->Add(StatusStr(validity, revoked, expired, disabled));
      else goto exit;

   // 3 bits  lvBits
   err6 = PGPGetKeyNumber(KeyRef, kPGPKeyPropBits, &bits);
   if (IsntPGPError(err6)) Key->Add(IntToStr(bits));
      else goto exit;

   // 4 algorithm  lvAlgo
   err6 = PGPGetKeyNumber(KeyRef, kPGPKeyPropAlgID, &algo);
   if (IsntPGPError(err6)) Key->Add(AlgorithmStr(algo));
      else goto exit;

   // 5 adks  lvADKs
   err6 = PGPCountAdditionalRecipientRequests(KeyRef, &adks);
   if (IsntPGPError(err6)) Key->Add(adks > 0 ? IntToStr(adks) : AnsiString(""));
      else goto exit;

   // 6 secret  lvSecret
   err6 = PGPGetKeyBoolean(KeyRef, kPGPKeyPropIsSecret, &secret);
   if (IsntPGPError(err6)) Key->Add(secret);
      else goto exit;

exit:

   if (IsPgpError(err6, "I was unable to get the requested key."));

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetSubKey(PGPKeyIterRef keyDBIter, TStrings *Key)
{
   assert(keyDBIter != NULL && Key != NULL && Key->Count == lvSecret + 1);

   PGPInt32 bits, algo;
   PGPSubKeyRef subKey = kInvalidPGPSubKeyRef;

   err6 = PGPKeyIterSubKey(keyDBIter, &subKey);
   if (err6 == kPGPError_EndOfIteration)
      goto exit;

   err6 = PGPGetSubKeyNumber(subKey, kPGPKeyPropBits, &bits);
   if (IsPGPError(err6))
      goto exit;

   Key->Strings[lvBits] = IntToStr(bits) + "/" + Key->Strings[lvBits];

   err6 = PGPGetSubKeyNumber(subKey, kPGPKeyPropAlgID, &algo);
   if (IsPGPError(err6))
      goto exit;

   Key->Strings[lvAlgo] = AlgorithmStr(algo) + "/" + Key->Strings[lvAlgo];

exit:

   if (err6 == kPGPError_EndOfIteration)
      err6 = kPGPError_NoErr;

   if (IsPgpError(err6, "I was unable to get the requested subkey!"));

   return err6;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKeyComplete(PGPKeyRef theKey, TStrings *Key)
{
   assert(PGPKeyRefIsValid(theKey) && Key != NULL);

   char userid[kPGPMaxUserIDSize];
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPBoolean revoked, expired, disabled, secret;
   PGPInt32 validity, bits, algo, version;
   PGPSize useridSize;
   PGPUInt32 adks;
   PGPKeyID keyID;
   Key->Clear();

   // userid
   err6 = PGPGetPrimaryUserIDNameBuffer(theKey, kPGPMaxUserIDSize - 1, userid, &useridSize);
   if (IsntPGPError(err6)) Key->Add("userid=" + AnsiString(userid));
      else goto exit;

   // keyid
   err6 = PGPGetKeyIDFromKey(theKey, &keyID);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
   if (IsntPGPError(err6)) Key->Add("keyid=" + AnsiString(keyidstr));
      else goto exit;

   // TODO: 0 GetKeyText need to get key version
   // version
   //PGPGetKeyNumber(theKey, kPGPKeyPropertyVersion, &version);
   //if (IsntPGPError(err6)) Key->Add("version=" + IntToStr(version));
   //   else goto exit;
   version = 0;
   Key->Add("version=" + IntToStr(version));

   // bits
   err6 = PGPGetKeyNumber(theKey, kPGPKeyPropBits, &bits);
   if (IsntPGPError(err6)) Key->Add("bits=" + IntToStr(bits));
      else goto exit;

   // validity
   err6 = PGPGetKeyNumber(theKey, kPGPKeyPropValidity, &validity);
   if (IsntPGPError(err6)) Key->Add("validity=" + IntToStr(validity));
      else goto exit;

   // algorithm
   err6 = PGPGetKeyNumber(theKey, kPGPKeyPropAlgID, &algo);
   if (IsntPGPError(err6)) Key->Add("algorithm=" + IntToStr(algo));
      else goto exit;

   // adks
   err6 = PGPCountAdditionalRecipientRequests(theKey, &adks);
   if (IsntPGPError(err6)) Key->Add("adks=" + IntToStr(adks));
      else goto exit;

   // secret
   err6 = PGPGetKeyBoolean(theKey, kPGPKeyPropIsSecret, &secret);
   if (IsntPGPError(err6)) Key->Add("secret=" + IntToStr(secret));
      else goto exit;

   // revoked
   err6 = PGPGetKeyBoolean(theKey, kPGPKeyPropIsRevoked, &revoked);
   if (IsntPGPError(err6)) Key->Add("revoked=" + IntToStr(revoked));
      else goto exit;

   // expired
   err6 = PGPGetKeyBoolean(theKey, kPGPKeyPropIsExpired, &expired);
   if (IsntPGPError(err6)) Key->Add("expired=" + IntToStr(expired));
      else goto exit;

   // disabled
   err6 = PGPGetKeyBoolean(theKey, kPGPKeyPropIsDisabled, &disabled);
   if (IsntPGPError(err6)) Key->Add("disabled=" + IntToStr(disabled));
      else goto exit;

exit:

   if (IsPgpError(err6, "I wasn't able to get the requested key"));

   return err6;
}

//---------------------------------------------------------------------------
// TPGPsdk6
//---------------------------------------------------------------------------

__fastcall TPGPsdk6::TPGPsdk6()
   : TPGPsdk()
{
}
//---------------------------------------------------------------------------

__fastcall TPGPsdk6::TPGPsdk6(TForm *Owner, const AnsiString &Pubring,
   const AnsiString &Secring, const AnsiString &Remring)
   : TPGPsdk(Owner, Pubring, Secring, Remring)
{
   //assert(FileExists(Pubring) && FileExists(Secring) && FileExists(Remring));

   Owner = Owner;

   SetCurrentDirectory(ExtractFilePath(Application->ExeName).c_str());
   InitSuccess = false;

   // TODO 0: Error Message for Fail LoadLibrary()
   hSdkLib6 = LoadLibrary("pgp_sdk.dll");
   if (hSdkLib6 == NULL)
      return;

   err6 = PGPsdkInit();
   if (IsPgpError(err6, "Unable to initialize PGPsdk!"))
      return;

   err6 = PGPGetSDKVersion(&sdkversion);
   if (IsPgpError(err6,"Failed to establish SDK version!"))
      return;

   err6 = PGPNewContext(kPGPsdkAPIVersion, &context);
   if (IsPgpError(err6,"Failed to initialize PGP context structure!"))
      return; // false;

   err6 = PGPsdkLoadDefaultPrefs(context);
   if (IsPgpError(err6,"Unable to load PGPsdk default preferences!"))
      return; // false;

   PGPFileSpecRef pubSpec = kInvalidPGPFileSpecRef;
   PGPFileSpecRef secSpec = kInvalidPGPFileSpecRef;

   // if first attempt fails, use prefSelector from earlier sdk version
   err6=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)5,&pubSpec);
   if (err6==-12000) err6=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)0,&pubSpec);
   if (IsPgpError(err6,"Unable to get PGP public keyring file spec!"))
      goto exit;

   char *filename;
   err6=PGPGetFullPathFromFileSpec(pubSpec, &filename);
   if (IsPgpError(err6,"Unable to get PGP public keyring path!"))
      goto exit;
   mainpubringpkr = AnsiString(filename).LowerCase();

   // if first attempt fails, use prefSelector from earlier sdk version
   err6=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)6,&secSpec);
   if (err6==-12000) err6=PGPsdkPrefGetFileSpec(context,(PGPsdkPrefSelector)1,&secSpec);
   if (IsPgpError(err6,"Unable to get PGP secret keyring file spec!"))
      goto exit;

   err6=PGPGetFullPathFromFileSpec(secSpec, &filename);
   if (IsPgpError(err6,"Unable to get PGP secret keyring path!"))
      goto exit;
   mainsecringskr = AnsiString(filename).LowerCase();

   err6=PGPGetSDKVersion(&apiversion);
   if (IsPgpError(err6,"Failed to get PGPsdk API version!"))
      goto exit;

   InitSuccess = true;

exit:
   if (PGPFileSpecRefIsValid(secSpec)) PGPFreeFileSpec(secSpec);
   if (PGPFileSpecRefIsValid(pubSpec)) PGPFreeFileSpec(pubSpec);
}
//---------------------------------------------------------------------------

__fastcall TPGPsdk6::~TPGPsdk6()
{
   if (PGPContextRefIsValid(context))
      PGPFreeContext(context);

   if (InitSuccess)
      PGPsdkCleanup();

   if (hSdkLib6 != NULL)
      FreeLibrary(hSdkLib6);
}
//---------------------------------------------------------------------------

int __fastcall TPGPsdk6::IdNo()
{
   return 6;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPGPsdk6::GetLastErrorStr()
{
   char *buf = new char[256];

   PGPGetErrorString(err6, 255, buf);
   AnsiString errmsg = buf;

   delete buf;

   return errmsg;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::ExportDefaultKeys(TStrings *exports)
{
   assert(exports != NULL);

   exports->Clear();
   TStringList *revokes = new TStringList();
   TStringList *disables = new TStringList();
   PGPKeySetRef keyset = kInvalidPGPKeySetRef;

   err6=PGPOpenDefaultKeyRings(context,false,&keyset);
   if (IsPgpError(err6,"Failed to open default keyrings!"))
      goto exit;

   err6=GetRevokedKeyIds(keyset,revokes);
   if (IsPGPError(err6))
      goto exit;

   err6=GetDisableKeyIds(keyset,disables);
   if (IsPGPError(err6))
      goto exit;

   err6=ExportKeySet(keyset, kPGPExportFormat_Basic, true, exports); // true = include privates
   if (IsPGPError(err6))
      goto exit;

   if (!disables->Text.IsEmpty()) exports->Insert(0,"Disabled: "+disables->CommaText);
   if (!revokes->Text.IsEmpty()) exports->Insert(0,"Revoked: "+revokes->CommaText);

exit:

   if (PGPKeySetRefIsValid(keyset)) PGPFreeKeySet(keyset);
   delete disables;
   delete revokes;
   return IsntPGPError(err6);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::ExportDefaultKeysFast(TStrings *exports)
{
   assert(exports != NULL);

   exports->Clear();
   PGPKeySetRef keyset = kInvalidPGPKeySetRef;
   PGPKeySetRef nodisables = kInvalidPGPKeySetRef;
   PGPFilterRef disfilter = kInvalidPGPFilterRef;
   PGPKeySetRef norevokes = kInvalidPGPKeySetRef;
   PGPFilterRef revfilter = kInvalidPGPFilterRef;

   err6=PGPOpenDefaultKeyRings(context,false,&keyset);
   if (IsPgpError(err6,"Failed to open default keyrings!"))
      goto exit;

   err6 = PGPNewKeyRevokedFilter(context,false,&revfilter);
   if (IsPgpError(err6,"Failed to create revoked key filter."))
      goto exit;

   err6 = PGPFilterKeySet(keyset,revfilter,&norevokes);
   if (IsPgpError(err6,"Failed to get keys from revoked filter."))
      goto exit;

   err6 = PGPNewKeyDisabledFilter(context,false,&disfilter);
   if (IsPgpError(err6,"Failed to create disabled key filter."))
      goto exit;

   err6 = PGPFilterKeySet(norevokes,disfilter,&nodisables);
   if (IsPgpError(err6,"Failed to get keys from disabled filter."))
      goto exit;

   err6=ExportKeySet(nodisables,kPGPExportFormat_Basic,true,exports);
   if (IsPGPError(err6))                            // true = include privates
      goto exit;

exit:

   if (PGPFilterRefIsValid(revfilter))
      PGPFreeFilter(revfilter);

   if (PGPKeySetRefIsValid(norevokes))
      PGPFreeKeySet(norevokes);

   if (PGPFilterRefIsValid(disfilter))
      PGPFreeFilter(disfilter);

   if (PGPKeySetRefIsValid(nodisables))
      PGPFreeKeySet(nodisables);

   if (PGPKeySetRefIsValid(keyset))
      PGPFreeKeySet(keyset);

   return IsntPGPError(err6);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::UpdateQSKeyrings(TControl *ForHourGlass)
{
   // just in case the QS keyrings are the main PGP keyrings and somehow got here
   if (pubringpkr.LowerCase() == mainpubringpkr.LowerCase()) return true;
   if (secringskr.LowerCase() == mainsecringskr.LowerCase()) return true;

   PGPKeySetRef qsKeySet = kInvalidPGPKeySetRef;
   PGPKeySetRef mainKeySet = kInvalidPGPKeySetRef;

   OpenKeyRingsMutable(pubringpkr, secringskr, &qsKeySet);
   OpenKeyRings(mainpubringpkr, mainsecringskr, &mainKeySet);
   PGPAddKeys(mainKeySet, qsKeySet);
   PGPCommitKeyRingChanges(qsKeySet);

   if (PGPKeySetRefIsValid(mainKeySet))
      PGPFreeKeySet(mainKeySet);

   if (PGPKeySetRefIsValid(qsKeySet))
      PGPFreeKeySet(qsKeySet);

   return true;
}
//---------------------------------------------------------------------------

AnsiString __fastcall MessageDecryptVerifyLine(bool first, int decs, int sigs)
{
   AnsiString line = "*** "+s(first?"BEGIN":"END")+" PGP ";
   if (decs) line += "DECRYPTED";
   if (decs&&sigs) line += "/";
   if (sigs) line += "VERIFIED";
   line += " MESSAGE ***";
   return line;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::VerifyClearSigned(TStrings *Message)
{
   assert(Message != NULL);

   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"verifying...");
   if (Message->IndexOf(BeginPgpSignedMessage) < 0)
      return true;

   void *data=NULL; PGPSize size;
   PGPKeySetRef PubRingRef = kInvalidPGPKeySetRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   Tdecinfo6 decinfo = { 0 }; decinfo.siginfos = new TStringList();

   PopHeaders(Message);

   err6 = PGPNewFileSpecFromFullPath(context, pubringpkr.c_str(), &PubKeyRing);
   if(IsPgpError(err6,"Failed to get PGP public keyring file spec."))
      goto exit;

   err6 = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_None, PubKeyRing, &PubRingRef);
   if (IsPgpError(err6,"Failed to open PGP public keyring"))
      goto exit;

   err6 = PGPDecode(context,
      PGPOKeySetRef(context, PubRingRef),
      PGPOEventHandler(context, DecodeEventHandler, &decinfo),
      PGPOInputBuffer(context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOLastOption(context),
      0,0,0,0
   );
   if (IsPgpError(err6,"Failed to decrypt message."))
      goto exit;

   Message->SetText((char*)data);
   if (decinfo.decrypts || decinfo.sigs)
   {
      Message->Insert(0,MessageDecryptVerifyLine(1,decinfo.decrypts,decinfo.sigs));
      Message->Insert(1,"");
      for (int i = 0; i < decinfo.siginfos->Count; i++)
         Message->Insert(i, decinfo.siginfos->Strings[i]);
      Message->Add("");
      Message->Add(MessageDecryptVerifyLine(0, decinfo.decrypts, decinfo.sigs));
   }

   PushHeaders(Message);

exit:

   delete decinfo.siginfos;

   if (PGPFileSpecRefIsValid(PubKeyRing))
      PGPFreeFileSpec(PubKeyRing);

   if (PGPKeySetRefIsValid(PubRingRef))
      PGPFreeKeySet(PubRingRef);

   if (data) PGPFreeData(data);

   return IsntPGPError(err6);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::DecryptMessage(TStrings *Message, char *Passphrase,
   TStrings *DecKeyIds, bool recurse)
{
   assert(Message != NULL && DecKeyIds != NULL && DecKeyIds->Count > 0);

   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"decrypting...");
   int i,idx;
   PGPKeyID KeyID;
   AnsiString keyidtext8;
   void *data=NULL; PGPSize size;
   PGPKeyRef DecKey = kInvalidPGPKeyRef;
   PGPKeySetRef KeyRingRef = kInvalidPGPKeySetRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;
   Tdecinfo6 decinfo = { 0 }; decinfo.siginfos = new TStringList();
   TStringList *keyids = new TStringList();
   keyids->Text = DecKeyIds->Text; //MessageDecIdText8(Message);

   err6 = PGPNewFileSpecFromFullPath(context, pubringpkr.c_str(), &PubKeyRing);
   if(IsPgpError(err6, "Failed to get QS PGP public keyring file spec."))
      goto Exit;

   err6 = PGPNewFileSpecFromFullPath(context, secringskr.c_str(), &SecKeyRing);
   if(IsPgpError(err6, "Failed to get QS PGP secret keyring file spec."))
      goto Exit;

   err6 = PGPOpenKeyRingPair(context, false, PubKeyRing, SecKeyRing, &KeyRingRef);
   if (IsPgpError(err6, "Failed to open QS PGP keyrings"))
      goto Exit;

   for (i=0; i<keyids->Count; i++)
   {
      keyidtext8 = keyids->Strings[i];
      err6 = PGPGetKeyIDFromString(keyidtext8.c_str(),&KeyID);
      if (IsntPGPError(err6)) break;
   }
   if (IsPgpError(err6, "Unable to get the required decryption keyid."))
      goto Exit;

   err6 = PGPGetKeyByKeyID(KeyRingRef, &KeyID, kPGPPublicKeyAlgorithm_Invalid, &DecKey);
   if (IsPgpError(err6, "Unable to get the required decryption key."))
      goto Exit;

   if (!PGPPassphraseIsValid(DecKey,PGPOPassphrase(context,Passphrase),PGPOLastOption(context)))
   {
      MessageBox(Owner->Handle,
         "Sorry, You've entered a bad passphrase!", "PGP Decrypt Error", MB_OK);
      err6 = kPGPError_BadPassphrase;
      goto Exit;
   }

   err6 = PGPDecode(context,
      PGPOKeySetRef(context, KeyRingRef),
      PGPOPassphrase(context, Passphrase),
      PGPORecursivelyDecode(context, recurse),
      PGPOEventHandler(context, DecodeEventHandler, &decinfo),
      PGPOInputBuffer(context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOLastOption(context),
      0,0
   );
   if (IsPgpError(err6,"Failed to decrypt message."))
      goto Exit;

   Message->SetText((char*)data);
   if (decinfo.decrypts || decinfo.sigs)
   {
      // remove clearsig that sdk leaves for some reason
      if (decinfo.sigs && (idx=Message->IndexOf(BeginPgpSignedMessage))>=0)
      {
         // drop -----Begin Pgp... line and blank line next
         Message->Delete(idx);
         if (idx<Message->Count) Message->Delete(idx);
         // find -----Begin Pgp Sig...
         if ((idx=Message->IndexOf(BeginPgpSignature))>=0)
         {
            // drop blank before and rest of msg
            if (idx>0) idx--; while (idx<Message->Count)
               Message->Delete(idx);
         }
      }
      Message->Insert(0, MessageDecryptVerifyLine(1, decinfo.decrypts, decinfo.sigs));
      Message->Insert(1, "");
      for (i=0; i<decinfo.siginfos->Count; i++)
         Message->Insert(i, decinfo.siginfos->Strings[i]);
      Message->Add("");
      Message->Add(MessageDecryptVerifyLine(0, decinfo.decrypts, decinfo.sigs));
   }

Exit:

   delete decinfo.siginfos;
   if (PGPFileSpecRefIsValid(SecKeyRing))
      PGPFreeFileSpec(SecKeyRing);

   if (PGPFileSpecRefIsValid(PubKeyRing))
      PGPFreeFileSpec(PubKeyRing);

   if (PGPKeySetRefIsValid(KeyRingRef))
      PGPFreeKeySet(KeyRingRef);

   if (data) PGPFreeData(data);

   return IsntPGPError(err6);
}
//---------------------------------------------------------------------------

/*bool __fastcall TPGPsdk6::SignMessage(TStrings *Message, char *SignKeyid,
   char *Passphrase)
{
   assert(Message != NULL && SignKeyid != NULL && Passphrase != NULL);

   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"signing...");

   PGPKeyID KeyID;
   AnsiString line,msgstr;
   void *data=NULL; PGPSize size;
   PGPKeyRef SignKey = kInvalidPGPKeyRef;
   PGPKeySetRef SecRingRef = kInvalidPGPKeySetRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;
   PGPHashAlgorithm HashAlg;

   PopHeaders(Message);

   err6 = PGPNewFileSpecFromFullPath(context, secringskr.c_str(), &SecKeyRing);
   if(IsPgpError(err6, "Failed to get PGP secret keyring file spec."))
      goto Exit;

   err6 = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_None, SecKeyRing, &SecRingRef);
   if (IsPgpError(err6, "Failed to open PGP secret keyring"))
      goto Exit;

   err6 = PGPGetKeyIDFromString(SignKeyid,&KeyID);
   if (IsPgpError(err6, "Unable to get the required signing keyid."))
      goto Exit;

   err6 = PGPGetKeyByKeyID(SecRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&SignKey);
   if (IsPgpError(err6, "Unable to get the required signing key."))
      goto Exit;

   err6 = PGPGetHashAlgUsed(SignKey, &HashAlg);
   if (IsPgpError(err6, "Unable to get the signing key hash algorithm."))
      goto Exit;

   if (HashAlg == kPGPHashAlgorithm_Invalid) HashAlg = kPGPHashAlgorithm_SHA;

   err6 = PGPEncode(context,
      PGPOInputBuffer( context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOSignWithKey( context, SignKey,
         PGPOPassphrase( context, Passphrase ),
         PGPOLastOption( context ), 0),
      PGPOHashAlgorithm( context, HashAlg),
      PGPOClearSign( context, true ),
      PGPOVersionString( context, "N/A" ),
      PGPODataIsASCII( context, true ),
      PGPOLastOption( context ),
      0,0,0,0,0,0,0,0
   );
   if (IsPgpError(err6, "Failed to clearsign message."))
      goto Exit;

   Message->SetText((char*)data);

   PushHeaders(Message);

Exit:

   if (PGPFileSpecRefIsValid(SecKeyRing))
      PGPFreeFileSpec(SecKeyRing);

   if (PGPKeySetRefIsValid(SecRingRef))
      PGPFreeKeySet(SecRingRef);

   if (data) PGPFreeData(data);

   return IsntPGPError(err6);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::EncryptSignMessage(TStrings *Message, TStrings *EncKeyIds,
   char *SignKeyid, char *Passphrase)
{
   assert(Message != NULL && EncKeyIds != NULL && SignKeyid != NULL && Passphrase != NULL);

   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"encrypting...");
   int i;
   PGPKeyID KeyID;
   AnsiString line,keyid,msgstr;
   void *data=NULL; PGPSize size;
   PGPKeyRef EncKey = kInvalidPGPKeyRef;
   PGPKeyRef SignKey = kInvalidPGPKeyRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPFileSpecRef SecKeyRing = kInvalidPGPFileSpecRef;
   PGPKeySetRef TheRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef EncRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef SinRingRef = kInvalidPGPKeySetRef;
   PGPHashAlgorithm HashAlg;

   PopHeaders(Message);

   err6 = PGPNewFileSpecFromFullPath(context, pubringpkr.c_str(), &PubKeyRing);
   if(IsPgpError(err6,"Failed to get PGP public keyring file spec."))
      goto Exit;
   err6 = PGPNewFileSpecFromFullPath(context, secringskr.c_str(), &SecKeyRing);
   if(IsPgpError(err6,"Failed to get PGP secret keyring file spec."))
      goto Exit;

   err6 = PGPOpenKeyRingPair(context, kPGPKeyRingOpenFlags_None, PubKeyRing,
      SecKeyRing, &TheRingRef);

   if (IsPgpError(err6,"Failed to open QS PGP keyrings"))
      goto Exit;

   err6 = PGPNewKeySet(context, &EncRingRef);
   if (IsPgpError(err6, "Failed creating keyset of requested keyids"))
      goto Exit;

   i = -1; while(++i < EncKeyIds->Count)
   {
      keyid = EncKeyIds->Strings[i].SubString(3, 8);
      err6 = PGPGetKeyIDFromString(keyid.c_str(), &KeyID);
      if (IsPgpError(err6, "Unable to get the required encrypting keyid."))
         goto Exit;

      err6 = PGPGetKeyByKeyID(TheRingRef, &KeyID, kPGPPublicKeyAlgorithm_Invalid, &EncKey);
      if (IsPgpError(err6, "Unable to get the required encrypting key."))
         goto Exit;

      err6 = PGPNewSingletonKeySet(EncKey, &SinRingRef);
      if (IsPgpError(err6, "Failed creating singlton keyset for key " + keyid))
         goto Exit;

      PGPAddKeys(SinRingRef,EncRingRef);
      PGPFreeKeySet(SinRingRef);
      SinRingRef = kInvalidPGPKeySetRef;
   }

   err6 = PGPGetKeyIDFromString(SignKeyid,&KeyID);
   if (IsPgpError(err6,"Unable to get the required signing keyid."))
      goto Exit;

   err6 = PGPGetKeyByKeyID(TheRingRef,&KeyID,kPGPPublicKeyAlgorithm_Invalid,&SignKey);
   if (IsPgpError(err6,"Unable to get the required signing key."))
      goto Exit;

   err6 = PGPGetHashAlgUsed(SignKey,&HashAlg);
   if (IsPgpError(err6,"Unable to get the signing key hash algorithm."))
      goto Exit;

   err6 = PGPEncode( context,
      PGPOEncryptToKeySet( context, EncRingRef ),
      PGPOSignWithKey( context, SignKey,
         PGPOPassphrase( context, Passphrase ),
         PGPOLastOption( context ), 0),
      PGPOHashAlgorithm( context, HashAlg),
      PGPOInputBuffer( context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOVersionString( context, "N/A" ),
      PGPOArmorOutput( context, true),
      PGPODataIsASCII( context, true),
      PGPOCompression( context, true),
      PGPOLastOption( context ),
      0,0,0,0,0,0
   );

   if (IsPgpError(err6,"Failed to encrypt message"))
      goto Exit;

   Message->SetText((char*)data);

   PushHeaders(Message);

Exit:

   if (PGPFileSpecRefIsValid(PubKeyRing))
      PGPFreeFileSpec(PubKeyRing);

   if (PGPFileSpecRefIsValid(SecKeyRing))
      PGPFreeFileSpec(SecKeyRing);

   if (PGPKeySetRefIsValid(TheRingRef))
      PGPFreeKeySet(TheRingRef);

   if (PGPKeySetRefIsValid(EncRingRef))
      PGPFreeKeySet(EncRingRef);

   if (PGPKeySetRefIsValid(SinRingRef))
      PGPFreeKeySet(SinRingRef);

   return IsntPGPError(err6);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::EncryptMessage(TStrings *Message, TStrings *KeyIds)
{
   assert(Message != NULL && KeyIds != NULL);

   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"encrypting...");
   int i;
   PGPKeyID KeyID;
   AnsiString line,keyid,msgstr;
   void *data=NULL; PGPSize size;
   PGPKeyRef EncKey = kInvalidPGPKeyRef;
   PGPFileSpecRef PubKeyRing = kInvalidPGPFileSpecRef;
   PGPKeySetRef PubRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef EncRingRef = kInvalidPGPKeySetRef;
   PGPKeySetRef SinRingRef = kInvalidPGPKeySetRef;

   PopHeaders(Message);

   err6 = PGPNewFileSpecFromFullPath(context, pubringpkr.c_str(), &PubKeyRing);
   if(IsPgpError(err6,"Failed to get PGP public keyring file spec."))
      goto Exit;

   err6 = PGPOpenKeyRing(context, kPGPKeyRingOpenFlags_None, PubKeyRing, &PubRingRef);
   if (IsPgpError(err6,"Failed to open PGP public keyring"))
      goto Exit;

   err6 = PGPNewKeySet(context,&EncRingRef);
   if (IsPgpError(err6,"Failed creating keyset of requested keyids"))
      goto Exit;

   i = -1; while(++i < KeyIds->Count)
   {
      keyid = KeyIds->Strings[i].SubString(3,8);
      err6 = PGPGetKeyIDFromString(keyid.c_str(), &KeyID);
      if (IsPgpError(err6, "Unable to get the required encrypting keyid."))
         goto Exit;

      err6 = PGPGetKeyByKeyID(PubRingRef, &KeyID, kPGPPublicKeyAlgorithm_Invalid, &EncKey);
      if (IsPgpError(err6, "Unable to get the required encrypting key."))
         goto Exit;

      err6 = PGPNewSingletonKeySet(EncKey, &SinRingRef);
      if (IsPgpError(err6, "Failed creating singlton keyset for key " + keyid))
         goto Exit;

      PGPAddKeys(SinRingRef, EncRingRef);
      PGPFreeKeySet(SinRingRef);
      SinRingRef = kInvalidPGPKeySetRef;
   }

   err6 = PGPEncode( context,
      PGPOEncryptToKeySet( context, EncRingRef ),
      PGPOInputBuffer( context, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(context, &data, -1,&size),
      PGPOVersionString(context, "N/A" ),
      PGPOArmorOutput(context, true),
      PGPODataIsASCII(context, true),
      PGPOCompression(context, true),
      PGPOLastOption(context),
      0,0,0,0,0,0,0,0
   );

   if (IsPgpError(err6,"Failed to encrypt message"))
      goto Exit;

   Message->SetText((char*)data);

   PushHeaders(Message);

Exit:

   if (PGPFileSpecRefIsValid(PubKeyRing))
      PGPFreeFileSpec(PubKeyRing);

   if (PGPKeySetRefIsValid(PubRingRef))
      PGPFreeKeySet(PubRingRef);

   if (PGPKeySetRefIsValid(EncRingRef))
      PGPFreeKeySet(EncRingRef);

   if (PGPKeySetRefIsValid(SinRingRef))
      PGPFreeKeySet(SinRingRef);

   return IsntPGPError(err6);
}*/
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::GetKeys(const AnsiString &pubringpkr,
   const AnsiString &secringskr, TStrings *Keys)
{
   assert(FileExists(pubringpkr) && FileExists(secringskr) && Keys != NULL);

   err6 = kPGPError_NoErr;
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyListRef keyList = kInvalidPGPKeyListRef;
   PGPKeyRef theKey = kInvalidPGPKeyRef;
   TStringList *key = new TStringList();

   err6 = OpenKeyRings(pubringpkr, secringskr, &keySet);
   if (IsPgpError(err6, "Unable to open keyring"))
      goto exit;

   err6 = PGPOrderKeySet(keySet, kPGPUserIDOrdering, &keyList);
   if (IsPgpError(err6, "Unable to create orderd key list!"))
      goto exit;

   err6 = PGPNewKeyIter(keyList, &keyIter);
   if (IsPgpError(err6, "Unable to create keylist iterator!"))
      goto exit;

   while (PGPKeyIterNext(keyIter, &theKey) != kPGPError_EndOfIteration)
   {
      err6 = GetKey(theKey, key);
      if (IsPGPError(err6))
         goto exit;

      err6 = GetSubKey(keyIter, key);
      if (IsPGPError(err6))
         goto exit;

      UpdateDescriptionField(key);
      AddKeyToKeys(key, Keys);
   }

exit:

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeyListRefIsValid(keyList))
      PGPFreeKeyList(keyList);

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   delete key;

   return err6 == kPGPError_NoErr;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::GetKeys(const AnsiString &pubring, TStrings *Keys)
{
   assert(FileExists(pubringpkr) && Keys != NULL);

   err6 = kPGPError_NoErr;
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyListRef keyList = kInvalidPGPKeyListRef;
   PGPKeyRef theKey = kInvalidPGPKeyRef;
   TStringList *key = new TStringList();

   err6 = OpenKeyRing(pubring, &keySet);
   if (IsPgpError(err6, "Unable to open keyring!"))
      goto exit;

   err6 = PGPOrderKeySet(keySet, kPGPUserIDOrdering, &keyList);
   if (IsPgpError(err6, "Unable to create orderd key list!"))
      goto exit;

   err6 = PGPNewKeyIter(keyList, &keyIter);
   if (IsPgpError(err6, "Unable to create keylist iterator!"))
      goto exit;

   while (PGPKeyIterNext(keyIter, &theKey) != kPGPError_EndOfIteration)
   {
      err6 = GetKeyComplete(theKey, key);
      if (IsPGPError(err6))
         goto exit;

      err6 = GetSubKey(keyIter, key);
      if (IsPGPError(err6))
         goto exit;

      UpdateDescriptionField(key);
      AddKeyToKeys(key, Keys);
   }

exit:

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeyListRefIsValid(keyList))
      PGPFreeKeyList(keyList);

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   delete key;

   return err6 == kPGPError_NoErr;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::GetKeysLV(const AnsiString &pubring, const AnsiString &secring,
   TStrings *Keys)
{
   assert(FileExists(pubring) && FileExists(secring) && Keys != NULL);

   err6 = kPGPError_NoErr;
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyListRef keyList = kInvalidPGPKeyListRef;
   PGPKeyRef theKey = kInvalidPGPKeyRef;
   TStringList *key = new TStringList();

   err6 = OpenKeyRings(pubring, secring, &keySet);
   if (IsPgpError(err6, "Unable to open keyrings!"))
      goto exit;

   err6 = PGPOrderKeySet(keySet, kPGPUserIDOrdering, &keyList);
   if (IsPgpError(err6, "Unable to create orderd key list!"))
      goto exit;

   err6 = PGPNewKeyIter(keyList, &keyIter);
   if (IsPgpError(err6, "Unable to create keylist iterator!"))
      goto exit;

   while (PGPKeyIterNext(keyIter, &theKey) != kPGPError_EndOfIteration)
   {
      err6 = GetKey(theKey, key);
      if (IsPGPError(err6))
         goto exit;

      err6 = GetSubKey(keyIter, key);
      if (IsPGPError(err6))
         goto exit;

      UpdateDescriptionField(key);
      AddKeyToKeys(key, Keys);
   }

exit:

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeyListRefIsValid(keyList))
      PGPFreeKeyList(keyList);

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   delete key;

   return err6 == kPGPError_NoErr;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPGPsdk6::GetKeyID(const AnsiString &Pubring, const AnsiString &UserID)
{
   AnsiString keyid;
   TStringList *keyids = new TStringList;
   int x;

   err6 = ::GetKeyID(Pubring, UserID, keyids, preferredalgo);
   if (IsPGPError(err6))
      goto exit;

   assert(keyids->Count > 0);

   keyid = keyids->Strings[0];

   x = keyid.Pos(" ");
   if (x > 0)
      keyid = keyid.SubString(1, x - 1);

exit:

   delete keyids;
   return keyid;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::EncryptToUserID(const AnsiString &Pubring,
   const AnsiString &UserID, TStrings *Text)
{
   TStringList *userids = new TStringList();
   userids->Add(UserID);

   bool ok = EncryptToUserIDs(Pubring, userids, Text);

   delete userids;

   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::EncryptToUserIDs(const AnsiString &Pubring, TStrings *UserIDs,
   TStrings *Text)
{
   assert(FileExists(Pubring) && UserIDs != NULL && UserIDs->Count > 0 && Text != NULL);

   bool ok = false;
   TStringList *keyids = new TStringList();

   err6 = GetKeyIDs(Pubring, UserIDs, keyids, preferredalgo);
   if (IsPGPError(err6))
      goto exit;

   ok = EncryptTo(Pubring, keyids, Text);

exit:

   delete keyids;

   return IsntPGPError(err6) && ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::EncryptTo(const AnsiString &Pubring, TStrings *KeyIDs, TStrings *Text)
{
   assert(FileExists(Pubring) && KeyIDs != NULL && KeyIDs->Count > 0 && Text != NULL);

   err = ::EncryptText(Pubring, "", KeyIDs, "", "", Text);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::EncryptTo(const AnsiString &Pubring, const AnsiString &KeyID, TStrings *Text)
{
   assert(FileExists(Pubring) && KeyID != "" && Text != NULL);

   TStringList *keyids = new TStringList();
   keyids->Add(KeyID);

   bool ok = EncryptTo(Pubring, keyids, Text);

   delete keyids;

   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::GetKeyLV(const AnsiString &pubring, const AnsiString &secring,
   const AnsiString &keyid, AnsiString &keydesc)
{
   PGPKeyID keyID;
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyRef theKey = kInvalidPGPKeyRef;
   TStringList *key = new TStringList();

   err6 = OpenKeyRings(pubring, secring, &keySet);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyIDFromString(keyid.c_str(), &keyID);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPGetKeyByKeyID(keySet, &keyID, kPGPPublicKeyAlgorithm_Invalid, &theKey);
   if (IsPGPError(err6))
      goto exit;

   err6 = ::GetKey(theKey, key);
   if (IsPGPError(err6))
      goto exit;

   err6 = ::GetSubKey(theKey, key);
   if (IsPGPError(err6))
      goto exit;

   UpdateDescriptionField(key);

   for (int i = 0; i < key->Count; i++)
   {
      if (i > 0 && i < key->Count)
         keydesc += "\255";
      keydesc += key->Strings[i];
   }

exit:

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   delete key;

   return err == kPGPError_NoErr;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::EncryptTo(const AnsiString &Passphrase, TStrings *Text) // conventional
{
   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Encrypting...");
   AnsiString line, msgstr;
   void *data = NULL;
   PGPSize size;

   PopHeaders(Text);

   err = ::EncryptText(Passphrase, Text);

   PushHeaders(Text);

exit:

   if (data)
      PGPFreeData(data);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::ExportKey(const AnsiString pubring, const AnsiString secring,
   AnsiString KeyID, bool IncludePrivate, TStrings *Export)
{
   TStringList *KeyIDs = new TStringList();
   KeyIDs->Add(KeyID);

   bool ok = ExportKeys(pubring, secring, KeyIDs, IncludePrivate, Export);

   delete KeyIDs;
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::ExportKeys(const AnsiString pubring, const AnsiString secring,
   TStrings *KeyIDs, bool IncludePrivates, TStrings *Exports)
{
   PGPKeySetRef singleSet = kInvalidPGPKeySetRef;
   PGPKeySetRef exportSet = kInvalidPGPKeySetRef;
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyRef theKey = kInvalidPGPKeyRef;

   err6 = OpenKeyRings(pubring, secring, &keySet);
   if (IsPGPError(err6))
      goto exit;

   err6 = PGPNewKeySet(context, &exportSet);
   if (IsPGPError(err6))
      goto exit;

   for (int i = 0; i < KeyIDs->Count; i++)
   {
      err6 = GetKey(keySet, KeyIDs->Strings[i], &theKey, kPGPPublicKeyAlgorithm_Invalid);
      if (IsPGPError(err6))
         break;

      err6 = PGPNewSingletonKeySet(theKey, &singleSet);
      if (IsPGPError(err6))
         break;

      err6 = PGPAddKeys(singleSet, exportSet);
      if (IsPGPError(err6))
         break;

      if (PGPKeySetRefIsValid(singleSet))
         PGPFreeKeySet(singleSet);
   }

   if (IsPGPError(err6))
      goto exit;

   err = ExportKeySet(exportSet, kPGPExportFormat_Basic, IncludePrivates, Exports);
   if (IsPGPError(err6))
      goto exit;

exit:

   if (IsPgpError(err6, "Failed to export requested key!"));

   if (PGPKeySetRefIsValid(singleSet))
      PGPFreeKeySet(singleSet);

   if (PGPKeySetRefIsValid(exportSet))
      PGPFreeKeySet(singleSet);

   if (PGPKeySetRefIsValid(exportSet))
      PGPFreeKeySet(keySet);

   return IsntPGPError(err6);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::RandomMouseMoved()
{
   return PGPGlobalRandomPoolMouseMoved();
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk6::GetRandomBytes(void *dataBuf, int size)
{
   return PGPContextGetRandomBytes(context, dataBuf, size) != kPGPError_OutOfEntropy;
}
//---------------------------------------------------------------------------
