//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <registry.hpp>
#include "Global2.h"      // WM_SHOWSTATUS
#include "PGPsdk8.h"
#include "PGPsdk8Calls.h"

// higher level
PGPError __fastcall OpenKeyRings(const AnsiString &pubing, const AnsiString &secring, PGPKeyDBRef *keyDB);
PGPError __fastcall OpenKeyRingsMutable(const AnsiString &pubing, const AnsiString &secring, PGPKeyDBRef *keyDB);
PGPError __fastcall OpenKeyRings(const AnsiString &pubring, const AnsiString &secring,
   PGPOpenKeyDBFileOptions options, PGPKeyDBRef *keyDB);
PGPError __fastcall OpenKeyRing(const AnsiString &pubring, PGPKeyDBRef *keyDB);
PGPError __fastcall OpenImportKeyRing(const AnsiString &pubring, PGPKeyDBRef *keyDB);
// cipher
PGPError __fastcall EncryptText(const AnsiString &pubring, const AnsiString &secring, TStrings *keyIDs,
   const AnsiString &signKeyID, const AnsiString &signPassphrase, TStrings *Text);
PGPError __fastcall EncryptText(const AnsiString &Passphrase, TStrings *Text); // conventional
// key for list view
PGPError __fastcall GetKey(PGPKeyDBObjRef KeyRef, TStrings *Key);
PGPError __fastcall GetSubKey(PGPKeyIterRef keyDBIter, TStrings *Key);
PGPError __fastcall GetSubKey(PGPKeyDBObjRef theKey, TStrings *Key);
PGPError __fastcall GetKeyComplete(PGPKeyDBObjRef KeyRef, TStrings *Key);
PGPError __fastcall GetKey(PGPKeyDBRef keyDB, const AnsiString &KeyIDStr, PGPKeyDBObjRef *theKey,
   PGPPublicKeyAlgorithm PreferredAlgo = kPGPPublicKeyAlgorithm_Invalid);
AnsiString __fastcall GetPreferredKey(TStrings *keyids, const AnsiString &preferredalgo);
// keyids
PGPError __fastcall GetKeyID(const AnsiString &Pubring, const AnsiString &UserId, TStrings* KeyIDs,
   const AnsiString &PreferredAlgo);
PGPError __fastcall GetKeyIDs(const AnsiString &Pubring, TStrings* UserIds, TStrings* KeyIDs,
   const AnsiString &PreferredAlgo);
// miscellaneous
PGPError __fastcall ExportKeySet(PGPKeySetRef keys, PGPExportFormat exportFormat,
   bool IncludePrivates, TStrings *exports);
PGPError __fastcall GetDisableKeyIds(PGPKeySetRef keySet, TStrings *keyIDs);
PGPError __fastcall GetRevokedKeyIds(PGPKeySetRef keySet, TStrings *keyIDs);
void __fastcall AddKeyToKeys(TStrings *key, TStrings *keys);
void __fastcall UpdateDescriptionField(TStrings *key);
// support
AnsiString __fastcall PublicKeyAlgoStr(PGPPublicKeyAlgorithm pkalg);
AnsiString __fastcall StatusStr(int validity, PGPBoolean revoked, PGPBoolean expired, PGPBoolean disabled);
AnsiString __fastcall AlgorithmStr(int algo);
AnsiString __fastcall ValidityStr(int val);

//---------------------------------------------------------------------------
#pragma package(smart_init)
const AnsiString RegAppDataKey =
   "Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders";

//---------------------------------------------------------------------------
// Init PGP8
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRings(const AnsiString &pubring, const AnsiString &secring, PGPKeyDBRef *keyDB)
{
   assert(FileExists(pubring));

   if (secring.IsEmpty() && ExtractFileExt(pubring).LowerCase() == ".asc")
      return OpenImportKeyRing(pubring, keyDB);

   return OpenKeyRings(pubring, secring, kPGPOpenKeyDBFileOptions_None, keyDB);
}
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRingsMutable(const AnsiString &pubring, const AnsiString &secring,
   PGPKeyDBRef *keyDB)
{
   return OpenKeyRings(pubring, secring, kPGPOpenKeyDBFileOptions_Mutable, keyDB);
}
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRings(const AnsiString &pubring, const AnsiString &secring,
   PGPOpenKeyDBFileOptions options, PGPKeyDBRef *keyDB)
{
   PGPFileSpecRef pubFileSpec = kInvalidPGPFileSpecRef;
   PGPFileSpecRef secFileSpec = kInvalidPGPFileSpecRef;

   if (pubring != "")
   {
      err = PGPNewFileSpecFromFullPath(Context8, pubring.c_str(), &pubFileSpec);
      if(IsPgpError(err, "Unable to get PGP public keyring path!"))
         goto exit;
   }

   if (secring != "")
   {
      err = PGPNewFileSpecFromFullPath(Context8, secring.c_str(), &secFileSpec);
      if (IsPgpError(err, "Unable to get PGP secret keyring path!"))
         goto exit;
   }

   err = PGPOpenKeyDBFile(Context8, options, pubFileSpec, secFileSpec, keyDB);
   if (IsPgpError(err, "Unable to open PGP keyrings!"))
      goto exit;

exit:

   if (PGPFileSpecRefIsValid(pubFileSpec))
      PGPFreeFileSpec(pubFileSpec);

   if (PGPFileSpecRefIsValid(secFileSpec))
      PGPFreeFileSpec(secFileSpec);

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall OpenKeyRing(const AnsiString &pubring, PGPKeyDBRef *keyDB)
{
   assert(FileExists(pubring));

   if (ExtractFileExt(pubring).LowerCase() == ".asc")
      return OpenImportKeyRing(pubring, keyDB);

   return OpenKeyRings(pubring, "", keyDB);
}
//---------------------------------------------------------------------------

PGPError __fastcall OpenImportKeyRing(const AnsiString &pubring, PGPKeyDBRef *keyDB)
{

   err = kPGPError_NoErr;
   PGPFileSpecRef pubFileRef = kInvalidPGPFileSpecRef;

   err = PGPNewFileSpecFromFullPath(Context8, pubring.c_str(), &pubFileRef);
   if(IsPGPError(err))
      goto exit;

   err = PGPImport(Context8, keyDB, PGPOInputFile(Context8, pubFileRef),
      PGPOLastOption(Context8), 0, 0, 0, 0);
   if (IsPgpError(err, "OK"))
      goto exit;

exit:

   if (PGPFileSpecRefIsValid(pubFileRef))
      PGPFreeFileSpec(pubFileRef);

   return err;
}
//---------------------------------------------------------------------------

AnsiString __fastcall PublicKeyAlgoStr(PGPPublicKeyAlgorithm algo)
{
   switch (algo)
   {
      case kPGPPublicKeyAlgorithm_RSA:            return "RSA";
      case kPGPPublicKeyAlgorithm_RSAEncryptOnly: return "RSAE";
      case kPGPPublicKeyAlgorithm_RSASignOnly:    return "RSAS";
      case kPGPPublicKeyAlgorithm_ElGamal:        return "DH";
      case kPGPPublicKeyAlgorithm_DSA:            return "DSA";
      case kPGPPublicKeyAlgorithm_ECEncrypt:      return "ECE";
      case kPGPPublicKeyAlgorithm_ECSign:         return "ECS";
   }
   return "UNKNOWN";
}
//---------------------------------------------------------------------------

AnsiString __fastcall GetPreferredKey(TStrings *keyids, const AnsiString &preferredalgo)
{
 //changed this assertion to dialog (in GetKeyIDs below) explaining problem of no key found
 //assert(keyids->Count > 0);

   int index = 0;
   AnsiString key;
   for (int i = 0; i < keyids->Count; i++)
   {
      if (keyids->Strings[i].Pos(" " + preferredalgo))
      {
         index = i;
         break;
      }
   }

   if (keyids->Count > 0)
      key = keyids->Strings[index];

   return key;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKeyIDs(const AnsiString &Pubring, TStrings *UserIDs,
   TStrings* KeyIDs, const AnsiString &PreferredAlgo)
{
   assert(FileExists(Pubring) && UserIDs != NULL && UserIDs->Count > 0 && KeyIDs != NULL);

   KeyIDs->Clear();
   TStringList *keyids = new TStringList();
   AnsiString chosenkey;

   char keyIDstr[kPGPMaxKeyIDStringSize];
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyDBObjRef theKey = kInvalidPGPKeyDBObjRef;
   PGPFilterRef keyFilter = kInvalidPGPFilterRef;
   PGPPublicKeyAlgorithm pkalg;
   PGPKeyID keyID;

   err = OpenKeyRing(Pubring, &keyDB);
   if (IsPGPError(err))
      goto exit;

   for (int i = 0; i < UserIDs->Count; i++)
   {
      AnsiString userid = UserIDs->Strings[i];

      err = PGPNewKeyDBObjDataFilter(Context8, kPGPUserIDProperty_EmailAddress, userid.c_str(),
         userid.Length(), kPGPMatchCriterion_SubString, &keyFilter);
      if(IsPGPError(err))
         goto exit;

      err = PGPFilterKeyDB(keyDB, keyFilter, &keySet);
      if(IsPGPError(err))
         goto exit;

      err = PGPNewKeyIterFromKeySet(keySet, &keyIter);
      if (IsPGPError(err))
         goto exit;

      keyids->Clear();
      while (PGPKeyIterNextKeyDBObj(keyIter, kPGPKeyDBObjType_Key, &theKey) != kPGPError_EndOfIteration)
      {
         err = PGPGetKeyID(theKey, &keyID);
         if (IsPGPError(err))
            goto exit;

         err = PGPGetKeyIDAlgorithm(&keyID, &pkalg);
         if (IsPGPError(err))
            goto exit;

         err = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyIDstr);
         if (IsPGPError(err))
         goto exit;

         keyids->Add(AnsiString(keyIDstr) + " " + PublicKeyAlgoStr(pkalg));
      }

      // should be no empty keyids but check anyway
      int i = 0; while (i < keyids->Count)
      {
         if (keyids->Strings[i].Trim().IsEmpty())
            keyids->Delete(i);
         else
            i++;
      }

      chosenkey = GetPreferredKey(keyids, PreferredAlgo);
      if (chosenkey.IsEmpty())
      {
         AnsiString ErrMsg =
           "I have a problem. There does not seem to be PGP key(s) for:    " +
            UserIDs->Text + "\nI'm quitting so you can take a look.";

         Application->MessageBox(ErrMsg.c_str(), "Error", MB_OK);
         err = kPGPError_ItemWasDeleted;
         goto exit;
      }

      KeyIDs->Add(chosenkey);
    //KeyIDs->Add(GetPreferredKey(keyids, PreferredAlgo));

      if (PGPKeyIterRefIsValid(keyIter))
         PGPFreeKeyIter(keyIter);

      if (PGPKeySetRefIsValid(keySet))
         PGPFreeKeySet(keySet);

      if (PGPFilterRefIsValid(keyFilter))
         PGPFreeFilter(keyFilter);
   }

exit:

   if (PGPFilterRefIsValid(keyFilter))
      PGPFreeFilter(keyFilter);

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   delete keyids;

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKeyID(const AnsiString &Pubring, const AnsiString &UserID,
   TStrings* KeyIDs, const AnsiString &PreferredAlgo)
{
   assert(FileExists(Pubring) && UserID != "" && KeyIDs != NULL);

   TStringList *userids = new TStringList();
   userids->Add(UserID);

   err = GetKeyIDs(Pubring, userids, KeyIDs, PreferredAlgo);

   delete userids;

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKey(PGPKeyDBRef keyDB, const AnsiString &KeyIDStr, PGPKeyDBObjRef *theKey,
   PGPPublicKeyAlgorithm PreferredAlgo)
{
   assert(PGPKeyDBRefIsValid(keyDB) && !KeyIDStr.IsEmpty());

   PGPKeyID keyID;

   err = PGPNewKeyIDFromString(KeyIDStr.c_str(), PreferredAlgo, &keyID);
   if (IsPGPError(err))
      goto exit;

   err = PGPFindKeyByKeyID(keyDB, &keyID, theKey);
   if (IsPGPError(err))
      goto exit;

exit:

   if (IsPgpError(err, "Failed to retrieve the requested key."));

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKey(PGPKeyDBObjRef KeyRef, TStrings *Key)
{
   assert(KeyRef != NULL && Key != NULL);

   char userid[kPGPMaxUserIDSize];
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPBoolean revoked, expired, disabled, secret;
   PGPInt32 bits, validity, algo;
   PGPUInt32 adks, useridsize;
   PGPKeyID keyID;
   int version;

   Key->Clear();

   // 0 userid  lvUserId
   err = PGPGetPrimaryUserIDName(KeyRef, userid, kPGPMaxUserIDSize - 1, &useridsize);
   Key->Add(userid);

   // 1 keyid  lvKeyID
   err = PGPGetKeyID(KeyRef, &keyID);
   if (IsPGPError(err))
      goto exit;

   err = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
   if (IsPGPError(err))
      goto exit;

   Key->Add(keyidstr);

   // 2 status  lvStatus
   err = PGPGetKeyDBObjBooleanProperty(KeyRef, kPGPKeyProperty_IsRevoked, &revoked);
   if (IsPGPError(err))
      goto exit;

   err = PGPGetKeyDBObjBooleanProperty(KeyRef, kPGPKeyProperty_IsDisabled, &disabled);
   if (IsPGPError(err))
      goto exit;

   err = PGPGetKeyDBObjBooleanProperty(KeyRef, kPGPKeyProperty_IsExpired, &expired);
   if (IsPGPError(err))
      goto exit;

   err = PGPGetKeyDBObjNumericProperty(KeyRef, kPGPKeyProperty_Validity, &validity);
   if (IsntPGPError(err)) Key->Add(StatusStr(validity, revoked, expired, disabled));
      else goto exit;

   // 3 bits  lvBits
   err = PGPGetKeyDBObjNumericProperty(KeyRef, kPGPKeyProperty_Bits, &bits);
   if (IsntPGPError(err)) Key->Add(IntToStr(bits));
      else goto exit;

   // 4 algorithm  lvAlgo
   err = PGPGetKeyDBObjNumericProperty(KeyRef, kPGPKeyProperty_AlgorithmID, &algo);
   if (IsntPGPError(err)) Key->Add(AlgorithmStr(algo));
      else goto exit;

   // 5 adks  lvADKs
   err = PGPCountAdditionalRecipientRequests(KeyRef, &adks);
   if (IsntPGPError(err)) Key->Add(adks > 0 ? IntToStr(adks) : AnsiString(""));
      else goto exit;

   // 6 secret  lvSecret
   err = PGPGetKeyDBObjBooleanProperty(KeyRef, kPGPKeyProperty_IsSecret, &secret);
   if (IsntPGPError(err)) Key->Add(secret);
      else goto exit;

   // 7 version lvVersion
   err = PGPGetKeyDBObjNumericProperty(KeyRef, kPGPKeyProperty_Version, &version);
   if (IsntPGPError(err)) Key->Add(IntToStr(version));
      else goto exit;
exit:

   if (IsPgpError(err, "I was unable to get the requested key."));

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetSubKey(PGPKeyDBObjRef theKey, TStrings *Key)
{
   assert(theKey != NULL && Key != NULL);

   PGPInt32 bits, algo;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyDBObjRef subKey = kInvalidPGPKeyDBObjRef;
   PGPKeyID keyID;

   err = PGPNewOneKeySet(theKey, &keySet);
   if (IsPGPError(err))
      goto exit;

   err = PGPNewKeyIterFromKeySet(keySet, &keyIter);
   if (IsPGPError(err))
      goto exit;

   err = PGPKeyIterNextKeyDBObj(keyIter, kPGPKeyDBObjType_SubKey, &subKey);
   if (IsPGPError(err))
      goto exit;

   // keyid
   err = PGPGetKeyID(subKey, &keyID);
   if (IsPGPError(err))
      goto exit;

   err = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
   if (IsntPGPError(err)) Key->Add("subkeyid=" + AnsiString(keyidstr));
      else goto exit;

   // bits
   err = PGPGetKeyDBObjNumericProperty(subKey, kPGPSubKeyProperty_Bits, &bits);
   if (IsntPGPError(err)) Key->Add("subbits=" + IntToStr(bits));
      else goto exit;

   // algorithm
   err = PGPGetKeyDBObjNumericProperty(subKey, kPGPSubKeyProperty_AlgorithmID, &algo);
   if (IsntPGPError(err)) Key->Add("subalgo=" + IntToStr(algo));
      else goto exit;

exit:

   if (err == kPGPError_EndOfIteration)
      err = kPGPError_NoErr;

   if (IsPgpError(err, "Failed to get requested subkey!"));

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetSubKey(PGPKeyIterRef keyDBIter, TStrings *Key)
{
   assert(keyDBIter != NULL && Key != NULL);// && Key->Count == lvSecret + 1);

   PGPInt32 bits, algo;
   PGPKeyDBObjRef subKey = kInvalidPGPKeyDBObjRef;

   err = PGPKeyIterNextKeyDBObj(keyDBIter, kPGPKeyDBObjType_SubKey, &subKey);
   if (err == kPGPError_EndOfIteration)
      goto exit;

   err = PGPGetKeyDBObjNumericProperty(subKey, kPGPKeyProperty_Bits, &bits);
   if (IsPGPError(err))
      goto exit;

   Key->Strings[lvBits] = IntToStr(bits) + "/" + Key->Strings[lvBits];

   err = PGPGetKeyDBObjNumericProperty(subKey, kPGPKeyProperty_AlgorithmID, &algo);
   if (IsPGPError(err))
      goto exit;

   Key->Strings[lvAlgo] = AlgorithmStr(algo) + "/" + Key->Strings[lvAlgo];

exit:

   if (err == kPGPError_EndOfIteration)
      err = kPGPError_NoErr;

   if (IsPgpError(err, "I was unable to get the requested subkey!"));

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetKeyComplete(PGPKeyDBObjRef theKey, TStrings *Key)
{
   char userid[kPGPMaxUserIDSize];
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPBoolean revoked, expired, disabled, secret;
   PGPInt32 validity, bits, algo, version;
   PGPSize useridSize;
   PGPUInt32 adks;
   PGPKeyID keyID;
   Key->Clear();

   // userid
   err = PGPGetPrimaryUserIDName(theKey, userid, kPGPMaxUserIDSize - 1, &useridSize);
   if (IsntPGPError(err)) Key->Add("userid=" + AnsiString(userid));
      else goto exit;

   // keyid
   err = PGPGetKeyID(theKey, &keyID);
   if (IsPGPError(err))
      goto exit;

   err = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
   if (IsntPGPError(err)) Key->Add("keyid=" + AnsiString(keyidstr));
      else goto exit;

   // version
   PGPGetKeyDBObjNumericProperty(theKey, kPGPKeyProperty_Version, &version);
   if (IsntPGPError(err)) Key->Add("version=" + IntToStr(version));
      else goto exit;

   // bits
   err = PGPGetKeyDBObjNumericProperty(theKey, kPGPKeyProperty_Bits, &bits);
   if (IsntPGPError(err)) Key->Add("bits=" + IntToStr(bits));
      else goto exit;

   // validity
   err = PGPGetKeyDBObjNumericProperty(theKey, kPGPKeyProperty_Validity, &validity);
   if (IsntPGPError(err)) Key->Add("validity=" + IntToStr(validity));
      else goto exit;

   // algorithm
   err = PGPGetKeyDBObjNumericProperty(theKey, kPGPKeyProperty_AlgorithmID, &algo);
   if (IsntPGPError(err)) Key->Add("algorithm=" + IntToStr(algo));
      else goto exit;

   // adks
   err = PGPCountAdditionalRecipientRequests(theKey, &adks);
   if (IsntPGPError(err)) Key->Add("adks=" + IntToStr(adks));
      else goto exit;

   // secret
   err = PGPGetKeyDBObjBooleanProperty(theKey, kPGPKeyProperty_IsSecret, &secret);
   if (IsntPGPError(err)) Key->Add("secret=" + IntToStr(secret));
      else goto exit;

   // revoked
   err = PGPGetKeyDBObjBooleanProperty(theKey, kPGPKeyProperty_IsRevoked, &revoked);
   if (IsntPGPError(err)) Key->Add("revoked=" + IntToStr(revoked));
      else goto exit;

   // expired
   err = PGPGetKeyDBObjBooleanProperty(theKey, kPGPKeyProperty_IsExpired, &expired);
   if (IsntPGPError(err)) Key->Add("expired=" + IntToStr(expired));
      else goto exit;

   // disabled
   err = PGPGetKeyDBObjBooleanProperty(theKey, kPGPKeyProperty_IsDisabled, &disabled);
   if (IsntPGPError(err)) Key->Add("disabled=" + IntToStr(disabled));
      else goto exit;

exit:

   if (IsPgpError(err, "I wasn't able to get the requested key"));

   return err;
}

//---------------------------------------------------------------------------
// cipher
//---------------------------------------------------------------------------

PGPError __fastcall EncryptText(const AnsiString &Pubring, const AnsiString &Secring,
   TStrings *keyIDs, const AnsiString &signKeyID,
   const AnsiString &signPassphrase,
   TStrings *Text)
{
   assert(Text != NULL && keyIDs != NULL);

   PGPKeyID keyID;
   bool gotencrypt = false, gotsign = false;
   AnsiString line, keyid, msgstr;
   void *data = NULL; PGPSize size;
   PGPKeyDBRef secDB = kInvalidPGPKeyDBRef;
   PGPKeyDBRef pubDB = kInvalidPGPKeyDBRef;
   PGPKeyDBObjRef encKey = kInvalidPGPKeyDBObjRef;
   PGPKeyDBObjRef signKey = kInvalidPGPKeyDBObjRef;
   PGPKeySetRef encSet = kInvalidPGPKeySetRef;

   // create keyset and optionlist for text encryption
   if (!Pubring.IsEmpty() && keyIDs != NULL && keyIDs->Count > 0)
   {
      err = OpenKeyRing(Pubring, &pubDB);
      if (IsPGPError(err))
         goto exit;

      err = PGPNewEmptyKeySet(pubDB, &encSet);
      if (IsPGPError(err))
         goto exit;

      int i = 0; while(i < keyIDs->Count)
      {
         keyid = keyIDs->Strings[i].SubString(3,8);

         err = PGPNewKeyIDFromString(keyid.c_str(), kPGPPublicKeyAlgorithm_Invalid, &keyID);
         if (IsPGPError(err))
            goto exit;

         err = PGPFindKeyByKeyID(pubDB, &keyID, &encKey);
         if (IsPGPError(err))
            goto exit;

         PGPAddKey(encKey, encSet);
         i++;
      }

      gotencrypt = true;
   }

   // get key and create optionlist for signing
   if (!Secring.IsEmpty() && !signKeyID.IsEmpty())
   {
      err = OpenKeyRing(Secring, &secDB);
      if (IsPGPError(err))
         goto exit;

      keyid = signKeyID.SubString(3,8);

      err = PGPNewKeyIDFromString(keyid.c_str(), kPGPPublicKeyAlgorithm_Invalid, &keyID);
      if (IsPGPError(err))
            goto exit;

      err = PGPFindKeyByKeyID(secDB, &keyID, &signKey);
      if (IsPGPError(err))
         goto exit;

      gotsign = true;
   }

   if (gotencrypt && !gotsign)
   {
      err = PGPEncode( Context8,
         PGPOEncryptToKeySet(Context8, encSet),
         PGPOInputBuffer( Context8, Text->Text.c_str(), Text->Text.Length()),
         PGPOAllocatedOutputBuffer( Context8, &data, -1, &size),
         PGPOVersionString( Context8, "N/A" ),
         PGPOArmorOutput( Context8, true),
         PGPODataIsASCII( Context8, true),
         PGPOCompression( Context8, true),
         PGPOAskUserForEntropy(Context8, true),
         PGPOLastOption( Context8 ),
         0,0,0,0,0,0,0);
      if (IsPgpError(err, "Sorry, I failed to encrypt your text!"))
         goto exit;

      Text->SetText((char*)data);
      goto exit;
   }

   if (gotsign && !gotencrypt)
   {
      err = PGPEncode( Context8,
         PGPOSignWithKey( Context8, signKey,
            PGPOPassphrase( Context8, signPassphrase.c_str() ),
            PGPOLastOption( Context8 ),0,0),
         PGPOClearSign( Context8, true ),
         PGPOInputBuffer( Context8, Text->Text.c_str(), Text->Text.Length()),
         PGPOAllocatedOutputBuffer( Context8, &data, -1, &size),
         PGPOVersionString( Context8, "N/A" ),
         PGPOArmorOutput( Context8, true),
         PGPODataIsASCII( Context8, true),
         PGPOCompression( Context8, true),
         PGPOAskUserForEntropy(Context8, true),
         PGPOLastOption( Context8 ),
         0,0,0,0,0,0);
      if (IsPgpError(err, "Sorry, I failed to sign your text!"))
         goto exit;

      Text->SetText((char*)data);
      goto exit;
   }

   err = PGPEncode( Context8,
      PGPOEncryptToKeySet(Context8, encSet),
      PGPOSignWithKey( Context8, signKey,
         PGPOPassphrase( Context8, signPassphrase.c_str() ),
         PGPOLastOption( Context8 ),0,0),
      PGPOClearSign( Context8, false ),
      PGPOInputBuffer( Context8, Text->Text.c_str(), Text->Text.Length()),
      PGPOAllocatedOutputBuffer( Context8, &data, -1, &size),
      PGPOVersionString( Context8, "N/A" ),
      PGPOArmorOutput( Context8, true),
      PGPODataIsASCII( Context8, true),
      PGPOCompression( Context8, true),
      PGPOAskUserForEntropy(Context8, true),
      PGPOLastOption( Context8 ),
      0,0,0,0,0);
   if (IsPgpError(err, "Sorry, I failed to encrypt/sign your text!"))
      goto exit;

   Text->SetText((char*)data);

exit:

   if (PGPKeySetRefIsValid(encSet))
      PGPFreeKeySet(encSet);

   if (PGPKeyDBRefIsValid(pubDB))
      PGPFreeKeyDB(pubDB);

   if (PGPKeyDBRefIsValid(secDB))
      PGPFreeKeyDB(secDB);

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall EncryptText(const AnsiString &Passphrase, TStrings *Text)
{
   AnsiString line, msgstr;
   void *data = NULL;
   PGPSize size;
   err = PGPEncode( Context8,
      PGPOConventionalEncrypt( Context8,
         PGPOPassphrase( Context8, Passphrase.c_str() ),
         PGPOLastOption( Context8 ), 0),
      PGPOInputBuffer( Context8, Text->Text.c_str(), Text->Text.Length()),
      PGPOAllocatedOutputBuffer( Context8, &data, -1,&size),
      PGPOVersionString( Context8, "N/A" ),
      PGPOArmorOutput( Context8, true),
      PGPODataIsASCII( Context8, true),
      PGPOCompression( Context8, true),
      PGPOAskUserForEntropy(Context8, true),
      PGPOLastOption( Context8 ),
      0,0,0,0,0,0,0
   );

   if (IsPGPError(err))
      goto exit;

   Text->SetText((char*)data);

exit:

   if (data)
      PGPFreeData(data);
   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall ExportKeySet(PGPKeySetRef keys, PGPExportFormat exportFormat,
   bool IncludePrivates, TStrings *exports)
{
   void *data = NULL; PGPSize size;
   exports->Clear();

   err = PGPExport(Context8,
      PGPOExportKeySet(Context8, keys),
      PGPOAllocatedOutputBuffer(Context8, &data, -1, &size), // -1 = no size constraint
      PGPOExportFormat(Context8, exportFormat),
      PGPOExportPrivateKeys(Context8, IncludePrivates),
      PGPOVersionString(Context8, "N/A"),
      PGPOLastOption(Context8)
   );

   if (IsntPGPError(err))
      exports->SetText((char*)data);

   if (data)
      PGPFreeData(data);

   return err;
}

//---------------------------------------------------------------------------
// support functions
//---------------------------------------------------------------------------

AnsiString __fastcall AlgorithmStr(int algo)
{
   switch (algo)
   {
      case kPGPPublicKeyAlgorithm_RSA: return "RSA";
      case kPGPPublicKeyAlgorithm_RSAEncryptOnly: return "RSA(e)";
      case kPGPPublicKeyAlgorithm_RSASignOnly: return "RSA(s)";
      case kPGPPublicKeyAlgorithm_ElGamal: return "DH";
      case kPGPPublicKeyAlgorithm_DSA: return "DSS";
      case kPGPPublicKeyAlgorithm_ECEncrypt: return "EC(e)";
      case kPGPPublicKeyAlgorithm_ECSign: return "EC(s)";
   }
   return "unknown";
}
//---------------------------------------------------------------------------

AnsiString __fastcall ValidityStr(int val)
{
   switch (val)
   {
      case kPGPValidity_Unknown : return "";
      case kPGPValidity_Invalid : return "invalid";
      case kPGPValidity_Marginal: return "marginal";
      case kPGPValidity_Complete: return "ok";
   }
   return "unknown";
}
//---------------------------------------------------------------------------

AnsiString __fastcall StatusStr(int validity, PGPBoolean revoked,
   PGPBoolean expired, PGPBoolean disabled)
{
   if (revoked ) return "revoked";
   if (expired ) return "expired";
   if (disabled) return "disabled";

   return ValidityStr(validity);
}
//---------------------------------------------------------------------------

void __fastcall UpdateDescriptionField(TStrings *key)
{
   bool secret = key->Strings[lvSecret] == "1";
   AnsiString descrip = key->Strings[lvAlgo];

   if (descrip == "RSA") descrip += (secret ? " legacy key pair" : " legacy public key"); else
   if (descrip == "DH/DSS") descrip += (secret ? " key pair" : " public key"); else
   if (descrip == "RSA/RSA") descrip = AnsiString("RSA") +
      (secret ? " key pair" : " public key");

   if (descrip != key->Strings[lvAlgo])
      key->Strings[lvAlgo] = descrip;
}
//---------------------------------------------------------------------------

void __fastcall AddKeyToKeys(TStrings *key, TStrings *keys)
{
   AnsiString keystr;
   for (int i = 0; i < key->Count; i++)
   {
      if (i > 0 && i < key->Count)
         keystr += "\255";
      keystr += key->Strings[i];
   }

   keys->Add(keystr);
}
//---------------------------------------------------------------------------

PGPError __fastcall GetRevokedKeyIds(PGPKeySetRef keySet, TStrings *keyIDs)
{
   PGPKeyDBObjRef key = kInvalidPGPKeyDBObjRef;
   PGPFilterRef revFilt = kInvalidPGPFilterRef;
   PGPKeySetRef revKeys = kInvalidPGPKeySetRef;
   PGPKeyIterRef setIter = kInvalidPGPKeyIterRef;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeyID keyID;
   keyIDs->Clear();

   err = PGPNewKeyDBObjBooleanFilter(Context8, kPGPKeyProperty_IsRevoked, true, &revFilt);
   if (IsPgpError(err, "Failed to create revoked key filter."))
      goto exit;

   err = PGPFilterKeySet((PGPKeySetRef)keySet, revFilt, &revKeys);
   if (IsPgpError(err,"Failed to get revoked keys."))
      goto exit;

   err = PGPNewKeyIterFromKeySet(revKeys, &setIter);
   if (IsPgpError(err,"Failed to create key list iterator."))
      goto exit;

   while (IsntPGPError(err = PGPKeyIterNextKeyDBObj(setIter, kPGPKeyDBObjType_Key, &key)) && key)
   {
      err = PGPGetKeyID(key, &keyID);
      if (IsPgpError(err,"Failed to get key id from revoked key."))
         goto exit;

      err = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
      if (IsPgpError(err, "Failed to get key id string!"))
         goto exit;

      keyIDs->Add(keyidstr);
   }

   if (err == kPGPError_EndOfIteration) err = kPGPError_NoErr;
   if (IsPgpError(err,"Failed to complete iteration of revoked key list."))
      goto exit;

exit:
   if (PGPKeyIterRefIsValid(setIter))
      PGPFreeKeyIter(setIter);

   if (PGPKeySetRefIsValid(revKeys))
      PGPFreeKeySet(revKeys);

   if (PGPFilterRefIsValid(revFilt))
      PGPFreeFilter(revFilt);

   return err;
}
//---------------------------------------------------------------------------

PGPError __fastcall GetDisableKeyIds(PGPKeySetRef keySet, TStrings *keyIDs)
{
   PGPKeyDBObjRef key = kInvalidPGPKeyDBObjRef;
   PGPFilterRef disFilt = kInvalidPGPFilterRef;
   PGPKeySetRef disKeys = kInvalidPGPKeySetRef;
   PGPKeyIterRef setIter = kInvalidPGPKeyIterRef;
   char keyidstr[kPGPMaxKeyIDStringSize];
   PGPKeyID keyID;
   keyIDs->Clear();

   err = PGPNewKeyDBObjBooleanFilter(Context8, kPGPKeyProperty_IsDisabled, true, &disFilt);
   if (IsPgpError(err, "Failed to create disabled key filter."))
      goto exit;

   err = PGPFilterKeySet((PGPKeySetRef)keySet, disFilt, &disKeys);
   if (IsPgpError(err, "Failed to get disabled keys."))
      goto exit;

   err = PGPNewKeyIterFromKeySet(disKeys, &setIter);
   if (IsPgpError(err, "Failed to create key list iterator."))
      goto exit;

   while (IsntPGPError(err = PGPKeyIterNextKeyDBObj(setIter, kPGPKeyDBObjType_Key, &key)) && key)
   {
      err = PGPGetKeyID(key, &keyID);
      if (IsPgpError(err,"Failed to get key id from disabled key."))
         goto exit;

      err = PGPGetKeyIDString(&keyID, kPGPKeyIDString_Abbreviated, keyidstr);
      if (IsPgpError(err, "Failed to get key id string!"))
         goto exit;

      keyIDs->Add(keyidstr);
   }
   if (err == kPGPError_EndOfIteration) err = kPGPError_NoErr;
   if (IsPgpError(err, "Failed to complete iteration of revoked key list."))
      goto exit;

exit:

   if (PGPKeyIterRefIsValid(setIter))
      PGPFreeKeyIter(setIter);

   if (PGPKeySetRefIsValid(disKeys))
      PGPFreeKeySet(disKeys);

   if (PGPFilterRefIsValid(disFilt))
      PGPFreeFilter(disFilt);

   return err;
}

//---------------------------------------------------------------------------
// PGPsdk8
//---------------------------------------------------------------------------

__fastcall TPGPsdk8::TPGPsdk8(TForm *owner, unsigned options)
   : TPGPsdk()
{
   Owner = owner;
   err = kPGPError_NoErr;
   Context8 = kInvalidPGPContextRef;

   hSdkLib8 = LoadLibrary("pgpsdk.dll");
   if (hSdkLib8 == NULL)
   {
      initerrormsg = "Failed to load PGP library.";
      InitSuccess = false;
      return;
   }

   err = PGPsdkInit(options);
   if (IsPGPError(err))
   {
      initerrormsg = "Failed to initialize PGPsdk.";
      InitSuccess = false;
      return;
   }

   err = PGPNewContext(kPGPsdkAPIVersion, &Context8);
   if (IsPGPError(err))
   {
      initerrormsg = "Failed to initialize PGP context.";
      InitSuccess = false;
      return;
   }

   preferredalgo = "DSA";
   InitSuccess = true;
}
//---------------------------------------------------------------------------

__fastcall TPGPsdk8::TPGPsdk8(TForm *Owner, const AnsiString &Pubring,
   const AnsiString &Secring, const AnsiString &Remring)
   : TPGPsdk(Owner, Pubring, Secring, Remring)
{
   prefs = new TStringList();

   InitSuccess = false;

   hSdkLib8 = LoadLibrary("pgpsdk.dll");
   if (hSdkLib8 == NULL)
   {
      return;
   }
   err = PGPsdkInit(0);
   if (IsPgpError(err, "Unable to initialize PGPsdk!"))
      return;

   sdkversion = PGPGetPGPsdkVersion();
   if (sdkversion == (PGPUInt32)-1)
      return;

   apiversion = PGPGetPGPsdkAPIVersion();
   if (apiversion == (PGPUInt32)-1)
      return;

   err = PGPNewContext(kPGPsdkAPIVersion, &Context8);
   if (IsPgpError(err, "Failed to initialize PGP context structure!"))
      return;

   if (!LoadDefaultPrefs())
   {
      Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Failed PGP SDK initialization");
      return;
   }

   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Successfully initialized PGP SDK");

   preferredalgo = "";
   InitSuccess = true;
}
//---------------------------------------------------------------------------

__fastcall TPGPsdk8::~TPGPsdk8()
{
   delete prefs;

   if (PGPContextRefIsValid(Context8))
      PGPFreeContext(Context8);

   if (InitSuccess)
      PGPsdkCleanup();

   if (hSdkLib8 != NULL)
      FreeLibrary(hSdkLib8);
}
//---------------------------------------------------------------------------

int __fastcall TPGPsdk8::IdNo()
{
   return 8;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPGPsdk8::GetLastErrorStr()
{
   char *buf = new char[256];

   PGPGetErrorString(err, 255, buf);
   AnsiString errmsg = buf;

   delete buf;

   return errmsg;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::LoadDefaultPrefs()
{
   TRegistry *reg = new TRegistry();
   if (reg->OpenKey(RegAppDataKey, false))
      pgpprefstxt = reg->ReadString("AppData") + "\\PGP Corporation\\PGP\\PGPPrefs.txt";
   delete reg;

   if (!FileExists(pgpprefstxt))
      return false;

   int i, p;
   prefs->LoadFromFile(pgpprefstxt);
   i = 0; while (i < prefs->Count)
   {
      AnsiString pref = prefs->Strings[i].Trim();
      if (pref.LowerCase().Pos("privatekeyringfile") == 1 ||
          pref.LowerCase().Pos("publickeyringfile") == 1)
      {
         while (pref.SubString(pref.Length() - 1, 2) == " \\")
         {
            pref = pref.SubString(1, pref.Length() - 2).Trim();
            prefs->Delete(i);

            if (i == prefs->Count - 1)
               break;

            pref += prefs->Strings[i].Trim();
         }
         while ((p = pref.Pos("\"")) > 0)
            pref.Delete(p, 1);

         if (i < prefs->Count)
            prefs->Strings[i] = pref;
         else
            prefs->Add(pref);
      }
      i++;
   }

   mainsecringskr = prefs->Values["privatekeyringfile"].LowerCase();
   mainpubringpkr = prefs->Values["publickeyringfile"].LowerCase();
   return true;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::ExportDefaultKeys(TStrings *exports)
{
   TStringList *revokes = new TStringList();
   TStringList *disables = new TStringList();
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   exports->Clear();

   err = OpenKeyRings(mainpubringpkr, mainsecringskr, kPGPOpenKeyDBFileOptions_None, &keyDB);
   if (IsPgpError(err, "Failed to initialize PGP keyrings."))
      goto exit;

   err = PGPNewKeySet(keyDB, &keySet);
   if (IsPgpError(err, "Failed to initialize required PGP key set."))
      goto exit;

   err = GetRevokedKeyIds(keySet, revokes);
   if (IsPGPError(err))
      goto exit;

   err = GetDisableKeyIds(keySet, disables);
   if (IsPGPError(err))
      goto exit;

   err = ExportKeySet(keySet, kPGPExportFormat_Basic, true, exports); // true = include privates
   if (IsPGPError(err))
      goto exit;

   if (!disables->Text.IsEmpty())
      exports->Insert(0,"Disabled: "+disables->CommaText);

   if (!revokes->Text.IsEmpty())
      exports->Insert(0,"Revoked: "+revokes->CommaText);

exit:

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   delete disables;
   delete revokes;

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::ExportDefaultKeysFast(TStrings *exports)
{
   exports->Clear();
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPFilterRef revfilter = kInvalidPGPFilterRef;
   PGPKeySetRef norevokes = kInvalidPGPKeySetRef;
   PGPFilterRef disfilter = kInvalidPGPFilterRef;
   PGPKeySetRef nodisables = kInvalidPGPKeySetRef;

   err = OpenKeyRings(pubringpkr, secringskr, kPGPOpenKeyDBFileOptions_None, &keyDB);
   if (IsPgpError(err, "Failed to initialize PGP keyrings."))
      goto exit;

   err = PGPNewKeyDBObjBooleanFilter(Context8, kPGPKeyProperty_IsRevoked, false, &revfilter);
   if (IsPgpError(err, "Failed to create revoked key filter."))
      goto exit;

   err = PGPFilterKeyDB(keyDB, revfilter, &norevokes);
   if (IsPgpError(err, "Failed to get keys from revoked filter."))
      goto exit;

   err = PGPNewKeyDBObjBooleanFilter(Context8, kPGPKeyProperty_IsDisabled, false, &disfilter);
   if (IsPgpError(err, "Failed to create disabled key filter."))
      goto exit;

   err = PGPFilterKeySet(norevokes, disfilter, &nodisables);
   if (IsPgpError(err, "Failed to get keys from disabled filter."))
      goto exit;

   err = ExportKeySet(nodisables, kPGPExportFormat_Basic, true, exports); // true = include privates
   if (IsPgpError(err, "Failed to export keys!"))
      goto exit;

exit:

   if (PGPKeySetRefIsValid(nodisables))
      PGPFreeKeySet(nodisables);

   if (PGPFilterRefIsValid(disfilter))
      PGPFreeFilter(disfilter);

   if (PGPKeySetRefIsValid(norevokes))
      PGPFreeKeySet(norevokes);

   if (PGPFilterRefIsValid(revfilter))
      PGPFreeFilter(revfilter);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::UpdateQSKeyrings(TControl *ForHourGlass)
{
   // just in case the QS keyrings are the main PGP keyrings and somehow got here
   if (pubringpkr.LowerCase() == mainpubringpkr.LowerCase()) return true;
   if (secringskr.LowerCase() == mainsecringskr.LowerCase()) return true;

   PGPKeyDBRef qsKeyDB = kInvalidPGPKeyDBRef;
   PGPKeyDBRef mainKeyDB = kInvalidPGPKeyDBRef;
   PGPKeySetRef mainKeySet = kInvalidPGPKeySetRef;

   OpenKeyRings(mainpubringpkr, mainsecringskr, kPGPOpenKeyDBFileOptions_None, &mainKeyDB);
   OpenKeyRings(pubringpkr, secringskr, kPGPOpenKeyDBFileOptions_Mutable, &qsKeyDB);
   PGPNewKeySet(mainKeyDB, &mainKeySet);
   PGPCopyKeys(mainKeySet, qsKeyDB, NULL);
   // no commit needed

   if (PGPKeySetRefIsValid(mainKeySet))
      PGPFreeKeySet(mainKeySet);

   if (PGPKeyDBRefIsValid(mainKeyDB))
      PGPFreeKeyDB(mainKeyDB);

   if (PGPKeyDBRefIsValid(qsKeyDB))
      PGPFreeKeyDB(qsKeyDB);

   return false;
}

//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::VerifyClearSigned(TStrings *Message)
{
   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Verifying...");
   if (Message->IndexOf(BeginPgpSignedMessage) < 0)
      return true;

   int i;
   void *data = NULL; PGPSize size;
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   Tdecinfo8 decinfo = { 0 }; decinfo.siginfos = new TStringList();
   PopHeaders(Message);

   err = OpenKeyRings(pubringpkr, secringskr, kPGPOpenKeyDBFileOptions_None, &keyDB);
   if (IsPgpError(err, "Failed to initialize PGP keyrings."))
      goto exit;

   decinfo.keyDB = keyDB;

   err = PGPDecode(Context8,
      PGPOKeyDBRef(Context8, keyDB),
      PGPOEventHandler(Context8, DecodeEventHandler, &decinfo),
      PGPOInputBuffer(Context8, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(Context8, &data, -1, &size),
      PGPOLastOption(Context8),
      0,0,0,0
   );
   if (IsPgpError(err,"Failed to decrypt message."))
      goto exit;

   Message->SetText((char*)data);

   if (decinfo.decrypts || decinfo.sigs)
   {
      Message->Insert(0,MessageDecryptVerifyLine(1,decinfo.decrypts,decinfo.sigs));
      Message->Insert(1,"");
      for (i=0; i<decinfo.siginfos->Count; i++)
         Message->Insert(i,decinfo.siginfos->Strings[i]);
      Message->Add("");
      Message->Add(MessageDecryptVerifyLine(0,decinfo.decrypts,decinfo.sigs));
   }
   PushHeaders(Message);

exit:

   delete decinfo.siginfos;
   if (data)
      PGPFreeData(data);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::DecryptMessage(TStrings *Message, char *Passphrase,
   TStrings *DecKeyIds, bool recurse)
{
   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Decrypting...");
   bool ok;
   int i,idx;
   PGPKeyID keyID;
   AnsiString keyidtext8;
   void *data = NULL; PGPSize size;
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeyDBObjRef decKey = kInvalidPGPKeyDBObjRef;
   Tdecinfo8 decinfo = { 0 }; decinfo.siginfos = new TStringList();

   //PopHeaders(Message);

   err = OpenKeyRings(pubringpkr, secringskr, kPGPOpenKeyDBFileOptions_None, &keyDB);
   if (IsPgpError(err, "Failed to initialize PGP keyrings."))
      goto exit;

   ok = true;
   if (DecKeyIds != NULL && DecKeyIds->Count > 0) // NULL or count == 0 for verify clearsign
   {
      ok = false;
      for (i = 0; i < DecKeyIds->Count; i++)
      {
         keyidtext8 = DecKeyIds->Strings[i];

         err = PGPNewKeyIDFromString(keyidtext8.c_str(), kPGPPublicKeyAlgorithm_Invalid, &keyID);
         if (IsPGPError(err))
            continue;

         err = PGPFindKeyByKeyID(keyDB, &keyID, &decKey);
         if (IsPGPError(err))
            continue;

         if (PGPPassphraseIsValid(decKey, PGPOPassphrase(Context8, Passphrase),
            PGPOLastOption(Context8), 0, 0))
         {
            ok = true;
            break;
         }
      }
   }

   if (!ok)
   {
      MessageBox(Owner->Handle,
         "Sorry, You've entered a bad passphrase!", "PGP Decrypt Error", MB_OK);
      err = kPGPError_BadPassphrase;
      goto exit;
   }

   decinfo.keyDB = keyDB;

   err = PGPDecode(Context8,
      PGPOInputBuffer(Context8, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(Context8, &data, -1,&size),
      PGPOEventHandler(Context8, DecodeEventHandler, &decinfo),
      PGPORecursivelyDecode(Context8, recurse),
      PGPOKeyDBRef(Context8, keyDB),
      PGPOPassphrase(Context8, Passphrase),
      PGPOLastOption(Context8),
      0,0
   );

   if (IsPgpError(err, "Failed to decrypt message."))
      goto exit;

   Message->SetText((char*)data);

   if (decinfo.decrypts || decinfo.sigs)
   {
      // remove clearsig that sdk leaves for some reason
      if (decinfo.sigs && (idx=Message->IndexOf(BeginPgpSignedMessage))>=0)
      {
         // drop -----Begin Pgp... line and blank line next
         Message->Delete(idx);
         if (idx<Message->Count) Message->Delete(idx);
         // find -----Begin Pgp Sig...
         if ((idx=Message->IndexOf(BeginPgpSignature))>=0)
         {
            // drop blank before and rest of msg
            if (idx>0) idx--; while (idx<Message->Count)
               Message->Delete(idx);
         }
      }
      Message->Insert(0,MessageDecryptVerifyLine(1,decinfo.decrypts,decinfo.sigs));
      Message->Insert(1,"");
      for (i=0; i<decinfo.siginfos->Count; i++)
         Message->Insert(i,decinfo.siginfos->Strings[i]);
      Message->Add("");
      Message->Add(MessageDecryptVerifyLine(0,decinfo.decrypts,decinfo.sigs));
   }

   //PushHeaders(Message);

exit:

   delete decinfo.siginfos;

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   if (data)
      PGPFreeData(data);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

/*bool __fastcall TPGPsdk8::SignMessage(TStrings *Message, char *SignKeyid,
   char *Passphrase)
{
   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Signing...");
   PGPKeyID keyID;
   AnsiString line,msgstr;
   void *data = NULL; PGPSize size;
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeyDBObjRef sigKey = kInvalidPGPKeyDBObjRef;

   PopHeaders(Message);

   err = OpenKeyRings(pubringpkr, secringskr, kPGPOpenKeyDBFileOptions_None, &keyDB);
   if (IsPgpError(err, "Failed to open PGP keyrings"))
      goto Exit;

   err = PGPNewKeyIDFromString(SignKeyid, kPGPPublicKeyAlgorithm_Invalid, &keyID);
   if (IsPgpError(err, "Unable to get the required signing keyid."))
      goto Exit;

   err = PGPFindKeyByKeyID(keyDB, &keyID, &sigKey);
   if (IsPgpError(err, "Unable to get the required signing key."))
      goto Exit;

   err = PGPEncode(Context8,
      PGPOInputBuffer(Context8, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(Context8, &data, -1, &size),
      PGPOSignWithKey(Context8, sigKey,
         PGPOPassphrase(Context8, Passphrase ),
         PGPOLastOption(Context8 ),
         0,0),
      PGPOClearSign(Context8, true ),
      PGPOVersionString(Context8, "N/A" ),
      PGPODataIsASCII(Context8, true ),
      PGPOLastOption(Context8),
      0,0,0,0,0,0,0,0,0
   );
   if (IsPgpError(err,"Failed to clearsign message."))
      goto Exit;

   Message->SetText((char*)data);

   PushHeaders(Message);

Exit:

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   if (data)
      PGPFreeData(data);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::EncryptMessage(TStrings *Message, TStrings *KeyIds)
{
   assert(Message != NULL && KeyIds != NULL);

   //PopHeaders(Message);
   //EncryptText(Message, KeyIds);
   //PushHeaders(Message);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::EncryptSignMessage(TStrings *Message, TStrings *EncKeyIds,
   char *SigKeyId, char *Passphrase)
{
   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"EncryptSign w/PGP8...");
   int i;
   PGPKeyID keyID;
   AnsiString keyid;//,line, msgstr;
   void *data = NULL; PGPSize size;
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeyDBRef remDB = kInvalidPGPKeyDBRef;
   PGPKeySetRef encSet = kInvalidPGPKeySetRef;
   PGPKeyDBObjRef encKey = kInvalidPGPKeyDBObjRef;
   PGPKeyDBObjRef sigKey = kInvalidPGPKeyDBObjRef;
   PopHeaders(Message);

   err = OpenKeyRings(pubringpkr, secringskr, kPGPOpenKeyDBFileOptions_None, &keyDB);
   if (IsPgpError(err, "Failed to open PGP keyrings"))
      goto Exit;

   err = PGPNewEmptyKeySet(keyDB, &encSet);
   if (IsPgpError(err,"Failed creating keyset of requested keyids"))
      goto Exit;

   // get encryption keys
   i = 0; while (i < EncKeyIds->Count)
   {
      keyid = EncKeyIds->Strings[i].SubString(3,8);

      err = PGPNewKeyIDFromString(keyid.c_str(), kPGPPublicKeyAlgorithm_Invalid, &keyID);
      if (IsPgpError(err,"Unable to get the required encrypting keyid."))
         goto Exit;

      // get the key from the personal keyring
      err = PGPFindKeyByKeyID(keyDB, &keyID, &encKey);
      if (IsPGPError(err))
      {
         if (remDB == kInvalidPGPKeyDBRef)
         {
            // if not in the personal ring check pubring.asc
            err = OpenKeyRings(remringpkr, "", kPGPOpenKeyDBFileOptions_None, &remDB);
            if (IsPgpError(err, "Failed to open pubring.asc"))
               goto Exit;
         }

         err = PGPFindKeyByKeyID(remDB, &keyID, &encKey);
         if (IsPgpError(err, "Unable to find the required encrypting key: 0x" + keyid))
            goto Exit;
      }

      PGPAddKey(encKey, encSet);
      i++;
   }

   // get signing key
   err = PGPNewKeyIDFromString(SigKeyId, kPGPPublicKeyAlgorithm_Invalid, &keyID);
   if (IsPgpError(err, "Unable to get the required signing keyid."))
      goto Exit;

   err = PGPFindKeyByKeyID(keyDB, &keyID, &sigKey);
   if (IsPgpError(err,"Unable to get the required signing key."))
      goto Exit;

   // do the encrypt and sign
   err = PGPEncode( Context8,
      PGPOEncryptToKeySet( Context8, encSet ),
      PGPOSignWithKey( Context8, sigKey,
         PGPOPassphrase( Context8, Passphrase ),
         PGPOLastOption( Context8 ),
         0,0),
      PGPOInputBuffer( Context8, Message->Text.c_str(), Message->Text.Length()),
      PGPOAllocatedOutputBuffer(Context8, &data, -1,&size),
      PGPOVersionString( Context8, "N/A" ),
      PGPOArmorOutput( Context8, true),
      PGPODataIsASCII( Context8, true),
      PGPOCompression( Context8, true),
      PGPOLastOption( Context8 ),
      0,0,0,0,0,0,0
   );

   if (IsPgpError(err,"Failed to encrypt message"))
      goto Exit;

   Message->SetText((char*)data);
   PushHeaders(Message);

Exit:

   if (PGPKeySetRefIsValid(encSet))
      PGPFreeKeySet(encSet);

   if (PGPKeyDBRefIsValid(remDB))
      PGPFreeKeyDB(remDB);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   return IsntPGPError(err);
}*/
//---------------------------------------------------------------------------

AnsiString __fastcall TPGPsdk8::GetKeyID(const AnsiString &Pubring, const AnsiString &UserID)
{
   AnsiString keyid;
   TStringList *keyids = new TStringList;
   int x;

   err = ::GetKeyID(Pubring, UserID, keyids, preferredalgo);
   if (IsPGPError(err))
      goto exit;

   assert(keyids->Count > 0);

   keyid = keyids->Strings[0];

   x = keyid.Pos(" ");
   if (x > 0)
      keyid = keyid.SubString(1, x - 1);

exit:

   delete keyids;
   return keyid;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::EncryptToUserID(const AnsiString &Pubring,
   const AnsiString &UserID, TStrings *Text)
{
   TStringList *userids = new TStringList();
   userids->Add(UserID);

   bool ok = EncryptToUserIDs(Pubring, userids, Text);

   delete userids;

   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::EncryptToUserIDs(const AnsiString &Pubring, TStrings *UserIDs,
   TStrings *Text)
{
   assert(FileExists(Pubring) && UserIDs != NULL && UserIDs->Count > 0 && Text != NULL);

   bool ok = false;
   TStringList *keyids = new TStringList();

   err = GetKeyIDs(Pubring, UserIDs, keyids, preferredalgo);
   if (IsPGPError(err))
      goto exit;

   ok = EncryptTo(Pubring, keyids, Text);

exit:

   delete keyids;

   return IsntPGPError(err) && ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::EncryptTo(const AnsiString &Pubring, const AnsiString &KeyID, TStrings *Text)
{
   assert(FileExists(Pubring) && KeyID != "" && Text != NULL);

   TStringList *keyids = new TStringList();
   keyids->Add(KeyID);

   bool ok = EncryptTo(Pubring, keyids, Text);

   delete keyids;

   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::EncryptTo(const AnsiString &Pubring, TStrings *KeyIDs, TStrings *Text)
{
   assert(FileExists(Pubring) && KeyIDs != NULL && KeyIDs->Count > 0 && Text != NULL);

   err = EncryptText(Pubring, "", KeyIDs, "", "", Text);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::EncryptTo(const AnsiString &Passphrase, TStrings *Text) // conventional
{
   Owner->Perform(WM_SHOWSTATUS,0,(LPARAM)"Encrypting...");
   AnsiString line, msgstr;
   void *data = NULL;
   PGPSize size;

   PopHeaders(Text);

   err = EncryptText(Passphrase, Text);

   PushHeaders(Text);

exit:

   if (data)
      PGPFreeData(data);

   return IsntPGPError(err);
}

//---------------------------------------------------------------------------
// GetKeysListView
//---------------------------------------------------------------------------

//bool __fastcall TPGPsdk8::GetKeysLV(const AnsiString &pubring, TStrings *Keys)
//{
//   return GetKeysLV(pubring, "", Keys);
//}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::GetKeysLV(const AnsiString &pubring, const AnsiString &secring,
   TStrings *Keys)
{
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyDBObjRef theKey = kInvalidPGPKeyDBObjRef;
   TStringList *key = new TStringList();

   err = OpenKeyRings(pubring, secring, &keyDB);
   if (IsPGPError(err))
      goto exit;

   err = PGPNewKeyIterFromKeyDB(keyDB, &keyIter);
   if (IsPGPError(err))
      goto exit;

   while (PGPKeyIterNextKeyDBObj(keyIter, kPGPKeyDBObjType_Key, &theKey) !=
      kPGPError_EndOfIteration)
   {
      err = GetKey(theKey, key);
      if (IsPGPError(err))
         goto exit;

      err = GetSubKey(keyIter, key);
      if (IsPGPError(err))
         goto exit;

      UpdateDescriptionField(key);
      AddKeyToKeys(key, Keys);
   }

exit:

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   delete key;

   return err == kPGPError_NoErr;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::GetKeyLV(const AnsiString &pubring, const AnsiString &secring,
   const AnsiString &keyid, AnsiString &keydesc)
{
   PGPKeyID keyID;
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeyDBObjRef theKey = kInvalidPGPKeyDBObjRef;
   TStringList *key = new TStringList();

   err = OpenKeyRings(pubring, secring, &keyDB);
   if (IsPGPError(err))
      goto exit;

   err = PGPNewKeyIDFromString(keyid.c_str(), kPGPPublicKeyAlgorithm_Invalid, &keyID);
   if (IsPGPError(err))
      goto exit;

   err = PGPFindKeyByKeyID(keyDB, &keyID, &theKey);
   if (IsPGPError(err))
      goto exit;

   err = GetKey(theKey, key);
   if (IsPGPError(err))
      goto exit;

   err = GetSubKey(theKey, key);
   if (IsPGPError(err))
      goto exit;

   UpdateDescriptionField(key);

   for (int i = 0; i < key->Count; i++)
   {
      if (i > 0 && i < key->Count)
         keydesc += "\255";
      keydesc += key->Strings[i];
   }

exit:

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   delete key;

   return err == kPGPError_NoErr;
}

//---------------------------------------------------------------------------
// GetKeysFullInfo
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::GetKeys(const AnsiString &pubring, const AnsiString &secring,
   TStrings *Keys)
{
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyDBObjRef theKey = kInvalidPGPKeyDBObjRef;
   TStringList *key = new TStringList();

   err = OpenKeyRings(pubring, secring, &keyDB);
   if (IsPGPError(err))
      goto exit;

   err = PGPNewKeyIterFromKeyDB(keyDB, &keyIter);
   if (IsPgpError(err, "Unable to create keyring iterator!"))
      goto exit;

   while (PGPKeyIterNextKeyDBObj(keyIter, kPGPKeyDBObjType_Key, &theKey) !=
      kPGPError_EndOfIteration)
   {
      err = GetKeyComplete(theKey, key);
      if (IsPGPError(err))
         goto exit;

      err = GetSubKey(keyIter, key);
      if (IsPGPError(err))
         goto exit;

      UpdateDescriptionField(key);
      AddKeyToKeys(key, Keys);
   }

exit:

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   delete key;

   return err == kPGPError_NoErr;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::ExportKeys(const AnsiString Pubring, const AnsiString Secring,
   TStrings *KeyIDs, bool IncludePrivates, TStrings *Exports)
{
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeySetRef keySet = kInvalidPGPKeySetRef;
   PGPKeyDBObjRef theKey = kInvalidPGPKeyDBObjRef;

   err = OpenKeyRings(Pubring, Secring, &keyDB);
   if (IsPGPError(err))
      goto exit;

   err = PGPNewEmptyKeySet(keyDB, &keySet);
   if (IsPGPError(err))
      goto exit;

   for (int i = 0; i < KeyIDs->Count; i++)
   {
      err = GetKey(keyDB, KeyIDs->Strings[i], &theKey, kPGPPublicKeyAlgorithm_Invalid);
      if (IsPGPError(err))
         break;

      PGPAddKey(theKey, keySet);
      if (IsPGPError(err))
         break;
   }
   if (IsPGPError(err))
      goto exit;

   err = ExportKeySet(keySet, kPGPExportFormat_Basic, IncludePrivates, Exports);

exit:

   if (PGPKeySetRefIsValid(keySet))
      PGPFreeKeySet(keySet);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   return IsntPGPError(err);
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::ExportKey(const AnsiString Pubring, const AnsiString Secring,
   AnsiString KeyID, bool IncludePrivate, TStrings *Export)
{
   TStringList *KeyIDs = new TStringList();

   KeyIDs->Add(KeyID);
   bool ok = ExportKeys(Pubring, Secring, KeyIDs, IncludePrivate, Export);

   delete KeyIDs;
   return ok;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::GetKeys(const AnsiString &pubring, TStrings *Keys)
{
   PGPKeyDBRef keyDB = kInvalidPGPKeyDBRef;
   PGPKeyIterRef keyIter = kInvalidPGPKeyIterRef;
   PGPKeyDBObjRef theKey = kInvalidPGPKeyDBObjRef;
   TStringList *key = new TStringList();

   err = OpenKeyRing(pubring, &keyDB);
   if (IsPGPError(err))
      goto exit;

   err = PGPNewKeyIterFromKeyDB(keyDB, &keyIter);
   if (IsPGPError(err))
      goto exit;

   while (PGPKeyIterNextKeyDBObj(keyIter, kPGPKeyDBObjType_Key, &theKey) !=
      kPGPError_EndOfIteration)
   {
      err = GetKeyComplete(theKey, key);
      if (IsPGPError(err))
         goto exit;

      err = GetSubKey(keyIter, key);
      if (IsPGPError(err))
         goto exit;

      UpdateDescriptionField(key);
      AddKeyToKeys(key, Keys);
   }

exit:

   if (PGPKeyIterRefIsValid(keyIter))
      PGPFreeKeyIter(keyIter);

   if (PGPKeyDBRefIsValid(keyDB))
      PGPFreeKeyDB(keyDB);

   delete key;

   return err == kPGPError_NoErr;
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::RandomMouseMoved()
{
   return PGPGlobalRandomPoolMouseMoved();
}
//---------------------------------------------------------------------------

bool __fastcall TPGPsdk8::GetRandomBytes(void *dataBuf, int size)
{
   return PGPContextGetRandomBytes(Context8, dataBuf, size) != kPGPError_OutOfEntropy;
}
//---------------------------------------------------------------------------














